# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import numpy as np
from scipy.sparse import csr_matrix, lil_matrix, diags
from pandapower.pf.makeYbus_facts import calc_y_svc_pu

SMALL_NUMBER = 1e-9


def create_J_modification_svc(J, svc_buses, pvpq, pq, pq_lookup, Vm, x_control, svc_x_l_pu, svc_x_cvar_pu,
                              nsvc_controllable, svc_controllable):

    in_pq = np.isin(svc_buses, pq)
    lpvpq = len(pvpq)
    lpq = len(pq)

    one = np.ones(shape=len(svc_buses), dtype=np.float64)

    y_svc = calc_y_svc_pu(x_control, svc_x_l_pu, svc_x_cvar_pu)
    q_svc = np.square(Vm[svc_buses]) * y_svc

    rows = np.array([], dtype=np.float64)
    cols = np.array([], dtype=np.float64)
    data = np.array([], dtype=np.float64)

    # J_m = np.zeros_like(J.toarray())
    # controllable = np.arange(nsvc)[svc_controllable]

    # # J_C_Q_u
    # J_C_Q_u = np.zeros(shape=(len(pq), len(pq)), dtype=np.float64)
    # J_C_Q_u[pq_lookup[svc_buses[in_pq]], pq_lookup[svc_buses[in_pq]]] = 2 * q_svc[in_pq]
    # # count pvpq rows and pvpq columns from top left
    # J_m[len(pvpq):len(pvpq) + len(pq), len(pvpq):len(pvpq) + len(pq)] = J_C_Q_u

    rows = np.r_[rows, lpvpq + pq_lookup[svc_buses[in_pq]]]
    cols = np.r_[cols, lpvpq + pq_lookup[svc_buses[in_pq]]]
    data = np.r_[data, 2 * q_svc[in_pq]]

    # # J_C_Q_c
    # if np.any(svc_controllable):
    #     J_C_Q_c = np.zeros(shape=(len(pq), nsvc), dtype=np.float64)
    #     values = 2 * np.abs(Vm[svc_buses]) ** 2 * (np.cos(2 * x_control) - 1) / (np.pi * svc_x_l_pu)
    #     J_C_Q_c[pq_lookup[svc_buses[in_pq & svc_controllable]], controllable] = values[in_pq & svc_controllable]
    #     # count pvpq rows and pvpq columns from top left
    #     J_m[len(pvpq):len(pvpq) + len(pq),
    #     len(pvpq) + len(pq):len(pvpq) + len(pq) + nsvc_controllable] = J_C_Q_c[:, controllable]

    if np.any(svc_controllable):
        jcqc_values = 2* np.square(np.abs(Vm[svc_buses])) * (np.cos(2 * x_control) - 1) / (np.pi * svc_x_l_pu)
        rows = np.r_[rows, lpvpq + pq_lookup[svc_buses[in_pq & svc_controllable]]]
        cols = np.r_[cols, lpvpq + lpq:lpvpq + lpq + nsvc_controllable]
        data = np.r_[data, jcqc_values[in_pq & svc_controllable]]

    # # J_C_C_u
    # # d(Ui - Ui,set)/d(Ui) = dUi/dUi = 1
    # if np.any(svc_controllable):
    #     J_C_C_u = np.zeros(shape=(nsvc, len(pq)), dtype=np.float64)
    #     J_C_C_u[controllable, pq_lookup[svc_buses[in_pq & controllable]]] = 1
    #     # count pvpq rows and pvpq columns from top left
    #     J_m[len(pvpq) + len(pq):len(pvpq) + len(pq) + nsvc_controllable,
    #     len(pvpq):len(pvpq) + len(pq)] = J_C_C_u[controllable, :]

    if np.any(svc_controllable):
        rows = np.r_[rows, lpvpq + lpq:lpvpq + lpq + nsvc_controllable]
        cols = np.r_[cols, lpvpq + pq_lookup[svc_buses[in_pq & svc_controllable]]]
        data = np.r_[data, one[in_pq & svc_controllable]]

    J_m = csr_matrix((data, (rows, cols)), shape=J.shape, dtype=np.float64)

    return J_m


def create_J_modification_tcsc(J, V, y_tcsc_pu, x_control, tcsc_controllable,
                               tcsc_x_l_pu, f, t, pvpq, pq, pvpq_lookup, pq_lookup, nsvc):
    # y_tcsc_pu = calc_y_svc_pu(x_control, tcsc_x_l_pu, tcsc_x_cvar_pu)
    # S_tcsc_pu = V * (Ybus_tcsc.conj() @ V.conj())
    dY_TCSC_dx = 2 * (np.cos(2 * x_control) - 1) / (np.pi * tcsc_x_l_pu)

    Vf = V[f]
    Vt = V[t]

    Vmf = np.abs(Vf)
    Vmt = np.abs(Vt)

    # S_Fii = Vf * np.conj(Ybus_tcsc.toarray()[f, f] * Vf)
    # S_Fkk = Vt * np.conj(Ybus_tcsc.toarray()[t, t] * Vt)
    S_Fii = Vf * np.conj(y_tcsc_pu * Vf)
    S_Fkk = Vt * np.conj(y_tcsc_pu * Vt)
    P_Fii, Q_Fii = S_Fii.real, S_Fii.imag
    P_Fkk, Q_Fkk = S_Fkk.real, S_Fkk.imag

    # S_Fik = Vf * np.conj(Ybus_tcsc.toarray()[f, t] * Vt)
    # S_Fki = Vt * np.conj(Ybus_tcsc.toarray()[t, f] * Vf)
    S_Fik = Vf * np.conj(-y_tcsc_pu * Vt)
    S_Fki = Vt * np.conj(-y_tcsc_pu * Vf)
    P_Fik, Q_Fik = S_Fik.real, S_Fik.imag
    P_Fki, Q_Fki = S_Fki.real, S_Fki.imag


    # seems like it is not used:
    # S_ii = np.abs(V[f]) ** 2 * np.abs(Ybus[f, f]) * np.exp(1j * np.angle(Ybus[f, f].conj()))  ####
    # S_kk = np.abs(V[t]) ** 2 * np.abs(Ybus[t, t]) * np.exp(1j * np.angle(Ybus[t, t].conj()))  ####
    #
    # S_ij = Sbus[f] - S_ii
    # S_kj = Sbus[t] - S_kk

    S_Fi_dx = -1j * dY_TCSC_dx / y_tcsc_pu * (S_Fii + S_Fik)  # ybus_tcsc_pu already has -1j but we want it without -1j
    S_Fk_dx = -1j * dY_TCSC_dx / y_tcsc_pu * (S_Fkk + S_Fki)
    P_Fi_dx, Q_Fi_dx = S_Fi_dx.real, S_Fi_dx.imag
    P_Fk_dx, Q_Fk_dx = S_Fk_dx.real, S_Fk_dx.imag

    f_in_pq = np.isin(f, pq)
    t_in_pq = np.isin(t, pq)
    f_in_pvpq = np.isin(f, pvpq)
    t_in_pvpq = np.isin(t, pvpq)

    lpvpq = len(pvpq)
    lpq = len(pq)

    rows = np.array([], dtype=np.float64)
    cols = np.array([], dtype=np.float64)
    data = np.array([], dtype=np.float64)

    # J_C_P_d = np.zeros(shape=(len(pvpq), len(pvpq)), dtype=np.float64)
    # if np.any(f_in_pvpq):
    #     J_C_P_d[pvpq_lookup[f[f_in_pvpq]], pvpq_lookup[f[f_in_pvpq]]] = -Q_Fik
    # if np.any(f_in_pvpq & t_in_pvpq):
    #     J_C_P_d[pvpq_lookup[f[f_in_pvpq]], pvpq_lookup[t[t_in_pvpq]]] = Q_Fik
    #     J_C_P_d[pvpq_lookup[t[t_in_pvpq]], pvpq_lookup[f[f_in_pvpq]]] = Q_Fki
    # if np.any(t_in_pvpq):
    #     J_C_P_d[pvpq_lookup[t[t_in_pvpq]], pvpq_lookup[t[t_in_pvpq]]] = -Q_Fki
    # # J_C_P_d = np.array([[-S_Fik.imag, S_Fik.imag],
    # #                     [S_Fki.imag, -S_Fki.imag]]).reshape(2, 2)  # todo: generalize shapes to work with many TCSC
    #
    # # J_C_P_u = np.array([[S_Fik.real / Vm[f], S_Fik.real / Vm[t]],
    # #                     [S_Fki.real / Vm[f], S_Fki.real / Vm[t]]]).reshape(2, 2)

    rows = np.r_[
        rows, pvpq_lookup[f[f_in_pvpq]], pvpq_lookup[f[f_in_pvpq & t_in_pvpq]],
        pvpq_lookup[t[t_in_pvpq & f_in_pvpq]], pvpq_lookup[t[t_in_pvpq]]]
    cols = np.r_[
        cols, pvpq_lookup[f[f_in_pvpq]], pvpq_lookup[t[f_in_pvpq & t_in_pvpq]],
        pvpq_lookup[f[t_in_pvpq & f_in_pvpq]], pvpq_lookup[t[t_in_pvpq]]]
    data = np.r_[
        data, -Q_Fik[f_in_pvpq], Q_Fik[f_in_pvpq & t_in_pvpq],
        Q_Fki[t_in_pvpq & f_in_pvpq], -Q_Fki[t_in_pvpq]]

    # J_C_P_u = np.zeros(shape=(len(pvpq), len(pq)), dtype=np.float64)
    # if np.any(f_in_pvpq & f_in_pq):
    #     J_C_P_u[pvpq_lookup[f[f_in_pvpq]], pq_lookup[f[f_in_pq]]] = P_Fik / Vmf
    # if np.any(f_in_pvpq & t_in_pq):
    #     J_C_P_u[pvpq_lookup[f[f_in_pvpq]], pq_lookup[t[t_in_pq]]] = P_Fik / Vmt
    # if np.any(t_in_pvpq & f_in_pq):
    #     J_C_P_u[pvpq_lookup[t[t_in_pvpq]], pq_lookup[f[f_in_pq]]] = P_Fki / Vmf
    # if np.any(t_in_pvpq & t_in_pq):
    #     J_C_P_u[pvpq_lookup[t[t_in_pvpq]], pq_lookup[t[t_in_pq]]] = P_Fki / Vmt

    rows = np.r_[
        rows, pvpq_lookup[f[f_in_pvpq & f_in_pq]], pvpq_lookup[f[f_in_pvpq & t_in_pq]],
        pvpq_lookup[t[t_in_pvpq & f_in_pq]], pvpq_lookup[t[t_in_pvpq & t_in_pq]]]
    cols = np.r_[
        cols, lpvpq + pq_lookup[f[f_in_pvpq & f_in_pq]], lpvpq + pq_lookup[t[f_in_pvpq & t_in_pq]],
        lpvpq + pq_lookup[f[t_in_pvpq & f_in_pq]], lpvpq + pq_lookup[t[t_in_pvpq & t_in_pq]]]
    data = np.r_[
        data, (P_Fik / Vmf)[f_in_pvpq & f_in_pq], (P_Fik / Vmt)[f_in_pvpq & t_in_pq],
        (P_Fki / Vmf)[t_in_pvpq & f_in_pq], (P_Fki / Vmt)[t_in_pvpq & t_in_pq]]

    # J_C_Q_d = np.zeros(shape=(len(pq), len(pvpq)), dtype=np.float64)
    # if np.any(f_in_pvpq & f_in_pq):
    #     J_C_Q_d[pq_lookup[f[f_in_pq]], pvpq_lookup[f[f_in_pvpq]]] = P_Fik
    # if np.any(f_in_pq & t_in_pvpq):
    #     J_C_Q_d[pq_lookup[f[f_in_pq]], pvpq_lookup[t[t_in_pvpq]]] = -P_Fik
    # if np.any(t_in_pq & f_in_pvpq):
    #     J_C_Q_d[pq_lookup[t[t_in_pq]], pvpq_lookup[f[f_in_pvpq]]] = -P_Fki
    # if np.any(t_in_pvpq & t_in_pq):
    #     J_C_Q_d[pq_lookup[t[t_in_pq]], pvpq_lookup[t[t_in_pvpq]]] = P_Fki
    # # J_C_Q_d = np.array([[S_Fik.real, -S_Fik.real],
    # #                     [-S_Fki.real, S_Fki.real]]).reshape(2, 2)

    rows = np.r_[
        rows, lpvpq + pq_lookup[f[f_in_pq & f_in_pvpq]], lpvpq + pq_lookup[f[f_in_pq & t_in_pvpq]],
        lpvpq + pq_lookup[t[t_in_pq & f_in_pvpq]], lpvpq + pq_lookup[t[t_in_pq & t_in_pvpq]]]
    cols = np.r_[
        cols, pvpq_lookup[f[f_in_pq & f_in_pvpq]], pvpq_lookup[t[f_in_pq & t_in_pvpq]],
        pvpq_lookup[f[t_in_pq & f_in_pvpq]], pvpq_lookup[t[t_in_pq & t_in_pvpq]]]
    data = np.r_[
        data, P_Fik[f_in_pq & f_in_pvpq], -P_Fik[f_in_pq & t_in_pvpq],
        -P_Fki[t_in_pq & f_in_pvpq], P_Fki[t_in_pq & t_in_pvpq]]

    # J_C_Q_u = np.zeros(shape=(len(pq), len(pq)), dtype=np.float64)
    # if np.any(f_in_pq):
    #     J_C_Q_u[pq_lookup[f[f_in_pq]], pq_lookup[f[f_in_pq]]] = (2 * Q_Fii + Q_Fik) / Vmf
    # if np.any(f_in_pq & t_in_pq):
    #     J_C_Q_u[pq_lookup[f[f_in_pq]], pq_lookup[t[t_in_pq]]] = Q_Fik / Vmt
    #     J_C_Q_u[pq_lookup[t[t_in_pq]], pq_lookup[f[f_in_pq]]] = Q_Fki / Vmf
    # if np.any(t_in_pq):
    #     J_C_Q_u[pq_lookup[t[t_in_pq]], pq_lookup[t[t_in_pq]]] = (2 * Q_Fkk + Q_Fki) / Vmt
    # # J_C_Q_u = np.array([[(2 * S_Fii.imag + S_Fik.imag) / Vm[f], S_Fik.imag / Vm[f]],
    # #                     [S_Fki.imag / Vm[t], (2 * S_Fkk.imag + S_Fki.imag) / Vm[f]]]).reshape(2, 2)

    rows = np.r_[
        rows, lpvpq + pq_lookup[f[f_in_pq]], lpvpq + pq_lookup[f[f_in_pq & t_in_pq]],
        lpvpq + pq_lookup[t[t_in_pq & f_in_pq]], lpvpq + pq_lookup[t[t_in_pq]]]
    cols = np.r_[
        cols, lpvpq + pq_lookup[f[f_in_pq]], lpvpq + pq_lookup[t[f_in_pq & t_in_pq]],
        lpvpq + pq_lookup[f[t_in_pq & f_in_pq]], lpvpq + pq_lookup[t[t_in_pq]]]
    data = np.r_[
        data, ((2 * Q_Fii + Q_Fik) / Vmf)[f_in_pq], (Q_Fik / Vmt)[f_in_pq & t_in_pq],
        (Q_Fki / Vmf)[t_in_pq & f_in_pq], ((2 * Q_Fkk + Q_Fki) / Vmt)[t_in_pq]]

    # J_C_P_c = np.zeros(shape=(len(pvpq), nsvc + ntcsc), dtype=np.float64)
    # if np.any(f_in_pvpq):
    #     J_C_P_c[pvpq_lookup[f[f_in_pvpq]], (nsvc + np.arange(ntcsc))[f_in_pvpq]] = S_Fi_dx.real
    # if np.any(t_in_pvpq):
    #     J_C_P_c[pvpq_lookup[t[t_in_pvpq]], (nsvc + np.arange(ntcsc))[t_in_pvpq]] = S_Fk_dx.real
    # # J_C_P_c = np.array([[S_Fi_dx.real], [S_Fk_dx.real]]).reshape(2, 1)
    rows = np.r_[
        rows, pvpq_lookup[f[f_in_pvpq & tcsc_controllable]], pvpq_lookup[t[t_in_pvpq & tcsc_controllable]]]
    cols = np.r_[
        cols, lpvpq + lpq + nsvc:lpvpq + lpq + nsvc + np.count_nonzero(f_in_pvpq & tcsc_controllable),
        lpvpq + lpq + nsvc:lpvpq + lpq + nsvc + np.count_nonzero(t_in_pvpq & tcsc_controllable)]
    data = np.r_[
        data, P_Fi_dx[f_in_pvpq & tcsc_controllable], P_Fk_dx[t_in_pvpq & tcsc_controllable]]

    # J_C_Q_c = np.zeros(shape=(len(pq), nsvc + ntcsc), dtype=np.float64)
    # if np.any(f_in_pq):
    #     J_C_Q_c[pq_lookup[f[f_in_pq]], (nsvc + np.arange(ntcsc))[f_in_pq]] = S_Fi_dx.imag
    # if np.any(t_in_pq):
    #     J_C_Q_c[pq_lookup[t[t_in_pq]], (nsvc + np.arange(ntcsc))[t_in_pq]] = S_Fk_dx.imag
    # # J_C_Q_c = np.array([[S_Fi_dx.imag], [S_Fk_dx.imag]]).reshape(2, 1)
    rows = np.r_[
        rows, lpvpq + pq_lookup[f[f_in_pq & tcsc_controllable]], lpvpq + pq_lookup[t[t_in_pq & tcsc_controllable]]]
    cols = np.r_[
        cols, lpvpq + lpq + nsvc:lpvpq + lpq + nsvc + np.count_nonzero(f_in_pq & tcsc_controllable),
        lpvpq + lpq + nsvc:lpvpq + lpq + nsvc + np.count_nonzero(t_in_pq & tcsc_controllable)]
    data = np.r_[
        data, Q_Fi_dx[f_in_pq & tcsc_controllable], Q_Fk_dx[t_in_pq & tcsc_controllable]]

    # the signs are opposite here for J_C_C_d, J_C_C_u, J_C_C_c and I don't know why
    # main mode of operation - set point for p_to_mw:
    # J_C_C_d = np.zeros(shape=(len(pvpq), len(pvpq)))
    # J_C_C_d = np.zeros(shape=(nsvc + ntcsc, len(pvpq)), dtype=np.float64)
    # if np.any(f_in_pvpq):
    #     J_C_C_d[(nsvc + np.arange(ntcsc))[f_in_pvpq], pvpq_lookup[f[f_in_pvpq]]] = Q_Fik
    # if np.any(t_in_pvpq):
    #     J_C_C_d[(nsvc + np.arange(ntcsc))[t_in_pvpq], pvpq_lookup[t[t_in_pvpq]]] = -Q_Fik
    rows = np.r_[
        rows, lpvpq + lpq + nsvc:lpvpq + lpq + nsvc + np.count_nonzero(f_in_pvpq & tcsc_controllable),
        lpvpq + lpq + nsvc:lpvpq + lpq + nsvc + np.count_nonzero(t_in_pvpq & tcsc_controllable)]
    cols = np.r_[
        cols, pvpq_lookup[f[f_in_pvpq & tcsc_controllable]], pvpq_lookup[t[t_in_pvpq & tcsc_controllable]]]
    data = np.r_[
        data, Q_Fik[f_in_pvpq & tcsc_controllable], -Q_Fik[t_in_pvpq & tcsc_controllable]]


    # J_C_C_u = np.zeros(shape=(nsvc + ntcsc, len(pq)), dtype=np.float64)
    # if np.any(f_in_pq):
    #     J_C_C_u[(nsvc + np.arange(ntcsc))[f_in_pq], pq_lookup[f[f_in_pq]]] = P_Fik / Vmf
    # if np.any(t_in_pq):
    #     J_C_C_u[(nsvc + np.arange(ntcsc))[t_in_pq], pq_lookup[t[t_in_pq]]] = P_Fik / Vmt
    # rows = np.r_[
    #     rows, lpvpq + lpq + nsvc + np.arange(ntcsc)[f_in_pq & tcsc_controllable],
    #     lpvpq + lpq + nsvc + np.arange(ntcsc)[t_in_pq & tcsc_controllable]]
    rows = np.r_[
        rows, lpvpq + lpq + nsvc:lpvpq + lpq + nsvc + np.count_nonzero(f_in_pq & tcsc_controllable),
        lpvpq + lpq + nsvc:lpvpq + lpq + nsvc + np.count_nonzero(t_in_pq & tcsc_controllable)]
    cols = np.r_[
        cols, lpvpq + pq_lookup[f[f_in_pq & tcsc_controllable]], lpvpq + pq_lookup[t[t_in_pq & tcsc_controllable]]]
    data = np.r_[data, (P_Fik / Vmf)[f_in_pq & tcsc_controllable], (P_Fik / Vmt)[t_in_pq & tcsc_controllable]]

    # J_C_C_c = np.zeros(shape=(nsvc + ntcsc, nsvc + ntcsc), dtype=np.float64)
    # J_C_C_c[np.r_[nsvc:nsvc + ntcsc], np.r_[nsvc:nsvc + ntcsc]] = -S_Fi_dx.real  # .flatten()?
    rows = np.r_[rows, lpvpq + lpq + nsvc:lpvpq + lpq + nsvc + np.count_nonzero(tcsc_controllable)]
    cols = np.r_[cols, lpvpq + lpq + nsvc:lpvpq + lpq + nsvc + np.count_nonzero(tcsc_controllable)]
    data = np.r_[data, -P_Fi_dx[tcsc_controllable]]

    # alternative mode of operation: for Vm at to bus (mismatch and setpoint also must be adjusted):
    # J_C_C_d = np.zeros(shape=(len(x_control), len(pvpq)), dtype=np.float64)
    # J_C_C_u = np.zeros(shape=(len(x_control), len(pq)), dtype=np.float64)
    # J_C_C_u[np.arange(len(x_control)), pq_lookup[t]] = 1
    # J_C_C_c = np.zeros((len(x_control), len(x_control)), dtype=np.float64)

    # # todo: implement the "relevant" array to be used as mask
    # # todo: adjust indices with position counters
    # if np.all(tcsc_controllable):
    #     J_m = np.vstack([np.hstack([J_C_P_d, J_C_P_u, J_C_P_c]),
    #                      np.hstack([J_C_Q_d, J_C_Q_u, J_C_Q_c]),
    #                      np.hstack([J_C_C_d, J_C_C_u, J_C_C_c])])
    # elif np.any(tcsc_controllable) or np.any(svc_controllable):
    #     relevant = np.r_[np.arange(nsvc), nsvc + np.arange(ntcsc)[tcsc_controllable]]
    #     J_m = np.vstack([np.hstack([J_C_P_d, J_C_P_u, J_C_P_c[:, relevant]]),
    #                      np.hstack([J_C_Q_d, J_C_Q_u, J_C_Q_c[:, relevant]]),
    #                      np.hstack([J_C_C_d[relevant, :], J_C_C_u[relevant, :],
    #                                 J_C_C_c[:, relevant][relevant, :]])])
    # else:
    #     J_m = np.vstack([np.hstack([J_C_P_d, J_C_P_u]),
    #                      np.hstack([J_C_Q_d, J_C_Q_u])])

    #J_m = csr_matrix(J_m)

    J_m = csr_matrix((data, (rows, cols)), shape=J.shape, dtype=np.float64)

    return J_m


def create_J_modification_ssc_vsc(J, V, Vm, y_pu, f, t, pvpq, pq, pvpq_lookup, pq_lookup, control_v, control_delta):
    """
    creates the modification Jacobian matrix for SSC (STATCOM)

    Parameters
    ----------
    Vm
    control_delta
    V
        array of np.complex128
    y_pu
        admittances of the elements
    f
        array of np.int64
    t
        array of np.int64
    pvpq
        array of np.int64
    pq
        array of np.int64
    pvpq_lookup
        array of np.int64
    pq_lookup
        array of np.int64

    Returns
    -------
    J_m
        scipy.sparse.csr_matrix

    """
    #

    Vf = V[f]

    Vt = V[t]

    Vmf = Vm[f]
    Vmt = Vm[t]

    S_Fii = Vf * np.conj(y_pu * Vf)
    S_Fkk = Vt * np.conj(y_pu * Vt)
    P_Fii, Q_Fii = S_Fii.real, S_Fii.imag
    P_Fkk, Q_Fkk = S_Fkk.real, S_Fkk.imag

    S_Fik = Vf * np.conj(-y_pu * Vt)
    S_Fki = Vt * np.conj(-y_pu * Vf)
    P_Fik, Q_Fik = S_Fik.real, S_Fik.imag
    P_Fki, Q_Fki = S_Fki.real, S_Fki.imag

    # seems like it is not used:
    # S_ii = np.abs(V[f]) ** 2 * np.abs(Ybus[f, f]) * np.exp(1j * np.angle(Ybus[f, f].conj()))  ####
    # S_kk = np.abs(V[t]) ** 2 * np.abs(Ybus[t, t]) * np.exp(1j * np.angle(Ybus[t, t].conj()))  ####
    #
    # S_ij = Sbus[f] - S_ii
    # S_kj = Sbus[t] - S_kk


    f_in_pq = np.isin(f, pq)
    f_in_pvpq = np.isin(f, pvpq)
    t_in_pq = np.isin(t, pq)
    t_in_pvpq = np.isin(t, pvpq)

    zero = SMALL_NUMBER * np.ones(shape=S_Fii.shape, dtype=np.float64)
    one = np.ones(shape=S_Fii.shape, dtype=np.float64)

    rows = np.array([], dtype=np.float64)
    cols = np.array([], dtype=np.float64)
    data = np.array([], dtype=np.float64)

    # # todo: use _sum_by_group what multiple elements start (or end) at the same bus?
    # # J_C_P_d = np.zeros(shape=(len(pvpq) + len(x_control), len(pvpq) + len(x_control)), dtype=np.float64)
    # # J_C_P_d = np.zeros(shape=(len(pvpq), len(pvpq)), dtype=np.float64)
    # #if np.any(f_in_pvpq):
    #     # J_C_P_d[pvpq_lookup[f[f_in_pvpq]], pvpq_lookup[f[f_in_pvpq]]] = -S_Fik.imag
    #     # # J_C_P_d[pvpq_lookup[f[f_in_pvpq]], pvpq_lookup[t[f_in_pvpq]]+ len(x_control)] = S_Fik.imag
    #     # J_C_P_d[pvpq_lookup[f[f_in_pvpq]], pvpq_lookup[t[f_in_pvpq]]] = S_Fik.imag
    #     #
    #     # # J_C_P_d[pvpq_lookup[t[f_in_pvpq]] + len(x_control), pvpq_lookup[f[f_in_pvpq]]] = S_Fki.imag
    #     # # J_C_P_d[pvpq_lookup[t[f_in_pvpq]] + len(x_control), pvpq_lookup[t[f_in_pvpq]]+ len(x_control)] = -S_Fki.imag
    #     #
    #     # J_C_P_d[pvpq_lookup[t[f_in_pvpq]] , pvpq_lookup[f[f_in_pvpq]]] = S_Fki.imag
    #     # J_C_P_d[pvpq_lookup[t[f_in_pvpq]] , pvpq_lookup[t[f_in_pvpq]]] = -S_Fki.imag
    #
    # if np.any(f_in_pvpq):
    #     J_m[pvpq_lookup[f[f_in_pvpq]], pvpq_lookup[f[f_in_pvpq]]] = -S_Fik.imag
    # if np.any(f_in_pvpq & t_in_pvpq):
    #     J_m[pvpq_lookup[f[f_in_pvpq & t_in_pvpq]], pvpq_lookup[t[f_in_pvpq & t_in_pvpq]]] = S_Fik.imag
    #     J_m[pvpq_lookup[t[f_in_pvpq & t_in_pvpq]], pvpq_lookup[f[f_in_pvpq & t_in_pvpq]]] = S_Fki.imag
    # if np.any(t_in_pvpq):
    #     J_m[pvpq_lookup[t[t_in_pvpq]], pvpq_lookup[t[t_in_pvpq]]] = -S_Fki.imag

    rows = np.r_[rows, pvpq_lookup[f[f_in_pvpq]], pvpq_lookup[f[f_in_pvpq & t_in_pvpq]],
                       pvpq_lookup[t[f_in_pvpq & t_in_pvpq]], pvpq_lookup[t[t_in_pvpq]]]
    cols = np.r_[cols, pvpq_lookup[f[f_in_pvpq]], pvpq_lookup[t[f_in_pvpq & t_in_pvpq]],
                       pvpq_lookup[f[f_in_pvpq & t_in_pvpq]], pvpq_lookup[t[t_in_pvpq]]]
    # todo
    # data = np.r_[data, -S_Fik.imag[f_in_pvpq], S_Fik.imag[f_in_pvpq & t_in_pvpq], S_Fki.imag[f_in_pvpq & t_in_pvpq], -S_Fki.imag[t_in_pvpq]]
    # data = np.r_[data, -S_Fik.imag[f_in_pvpq], S_Fik.imag[f_in_pvpq & t_in_pvpq], one[f_in_pvpq & t_in_pvpq], zero[t_in_pvpq]]

    data = np.r_[data, -Q_Fik[f_in_pvpq], Q_Fik[f_in_pvpq & t_in_pvpq],
                 np.where(control_delta[f_in_pvpq & t_in_pvpq], one[f_in_pvpq & t_in_pvpq], Q_Fki[f_in_pvpq & t_in_pvpq]),
                 np.where(control_delta[t_in_pvpq], zero[t_in_pvpq], -Q_Fki[t_in_pvpq])]
    # data = np.r_[data, -Q_Fik[f_in_pvpq], Q_Fik[f_in_pvpq & t_in_pvpq],
    #              np.where(control_delta[f_in_pvpq & t_in_pvpq], one[f_in_pvpq & t_in_pvpq], Q_Fik[f_in_pvpq & t_in_pvpq]),
    #              np.where(control_delta[t_in_pvpq], zero[t_in_pvpq], -Q_Fik[t_in_pvpq])]


    # # J_C_P_u = np.zeros(shape=(len(pvpq), len(pq)), dtype=np.float64)
    # # J_C_P_u = np.zeros(shape=(len(pvpq)+ len(x_control), len(pq)+ len(x_control)), dtype=np.float64)
    #
    # #if np.any(f_in_pvpq & f_in_pq):  ## TODO check if this conditon includes all cases, and check trough tests
    #     # J_C_P_u[pvpq_lookup[f[f_in_pvpq]], pq_lookup[f[f_in_pq]]] = (2 * S_Fii.real + S_Fik.real) / Vmf
    #     # J_C_P_u[pvpq_lookup[f[f_in_pvpq]], pq_lookup[t[f_in_pq]] ] = S_Fik.real/Vmt
    #     #
    #     # J_C_P_u[pvpq_lookup[t[f_in_pvpq]], pq_lookup[f[f_in_pq]]] = S_Fki.real/Vmf
    #     # J_C_P_u[pvpq_lookup[t[f_in_pvpq]], pq_lookup[t[f_in_pq]]] = (2 * S_Fkk.real + S_Fki.real) / Vmt
    #
    # if np.any(f_in_pvpq & f_in_pq):
    #     J_m[pvpq_lookup[f[f_in_pvpq]], len(pvpq)+pq_lookup[f[f_in_pq]]] = (2 * S_Fii.real + S_Fik.real) / Vmf
    # if np.any(f_in_pvpq & t_in_pq):
    #     J_m[pvpq_lookup[f[f_in_pvpq & t_in_pq]], len(pvpq)+pq_lookup[t[f_in_pvpq & t_in_pq]] ] = S_Fik.real/Vmt
    #     J_m[pvpq_lookup[t[f_in_pvpq & t_in_pq]], len(pvpq)+pq_lookup[f[f_in_pvpq & t_in_pq]]] = S_Fki.real/Vmf
    # if np.any(t_in_pvpq & t_in_pq):
    #     J_m[pvpq_lookup[t[t_in_pvpq & t_in_pq]], len(pvpq)+pq_lookup[t[t_in_pvpq & t_in_pq]]] = (2 * S_Fkk.real + S_Fki.real) / Vmt

    rows = np.r_[rows, pvpq_lookup[f[f_in_pvpq]], pvpq_lookup[f[f_in_pvpq & t_in_pq]], pvpq_lookup[t[f_in_pvpq & t_in_pq]], pvpq_lookup[t[t_in_pvpq & t_in_pq]]]
    cols = np.r_[cols, len(pvpq)+pq_lookup[f[f_in_pq]], len(pvpq)+pq_lookup[t[f_in_pvpq & t_in_pq]], len(pvpq)+pq_lookup[f[f_in_pvpq & t_in_pq]], len(pvpq)+pq_lookup[t[t_in_pvpq & t_in_pq]]]
    # data = np.r_[data, ((2 * S_Fii.real + S_Fik.real) / Vmf)[f_in_pvpq], (S_Fik.real/Vmt)[f_in_pvpq & t_in_pq], (S_Fki.real/Vmf)[f_in_pvpq & t_in_pq], ((2 * S_Fkk.real + S_Fki.real) / Vmt)[t_in_pvpq & t_in_pq]]

    data = np.r_[data, ((2 * P_Fii + P_Fik) / Vmf)[f_in_pvpq], (P_Fik/Vmt)[f_in_pvpq & t_in_pq],
                 np.where(control_delta[f_in_pvpq & t_in_pq], zero[f_in_pvpq & t_in_pq], (P_Fki / Vmf)[f_in_pvpq & t_in_pq]),
                 np.where(control_delta[t_in_pvpq & t_in_pq], zero[t_in_pvpq & t_in_pq], ((2 * P_Fkk + P_Fki) / Vmt)[t_in_pvpq & t_in_pq])]
    # data = np.r_[data, ((2 * P_Fii + P_Fik) / Vmf)[f_in_pvpq], (P_Fik / Vmt)[f_in_pvpq & t_in_pq],
    #              np.where(control_delta[f_in_pvpq & t_in_pq], zero[f_in_pvpq & t_in_pq], ((2 * P_Fii + P_Fik) / Vmf)[f_in_pvpq & t_in_pq]),
    #              np.where(control_delta[t_in_pvpq & t_in_pq], zero[t_in_pvpq & t_in_pq], (P_Fik / Vmt)[t_in_pvpq & t_in_pq])]


    # # J_C_Q_d = np.zeros(shape=(len(pq), len(pvpq)), dtype=np.float64)
    # # J_C_Q_d = np.zeros(shape=(len(pq)+ len(x_control), len(pvpq)+ len(x_control)), dtype=np.float64)
    # #if np.any(f_in_pvpq & f_in_pq):
    #     # J_C_Q_d[pq_lookup[f[f_in_pq]], pvpq_lookup[f[f_in_pvpq]]] = S_Fik.real
    #     # J_C_Q_d[pq_lookup[f[f_in_pq]], pvpq_lookup[t[f_in_pvpq]]] = -S_Fik.real
    #     #
    #     # # J_C_Q_d[pq_lookup[t[f_in_pq]]+ len(x_control), pvpq_lookup[f[f_in_pvpq]]] = 0
    #     # # J_C_Q_d[pq_lookup[t[f_in_pq]]+ len(x_control), pvpq_lookup[t[f_in_pvpq]]+ len(x_control)] = 0
    # if np.any(f_in_pvpq & f_in_pq):
    #     J_m[len(pvpq) + pq_lookup[f[f_in_pq]], pvpq_lookup[f[f_in_pvpq]]] = S_Fik.real  # todo: should it be & or |?
    # if np.any(f_in_pq & t_in_pvpq):
    #     J_m[len(pvpq) + pq_lookup[f[f_in_pq & t_in_pvpq]], pvpq_lookup[t[f_in_pq & t_in_pvpq]]] = -S_Fik.real
    # if np.any(f_in_pvpq & t_in_pq):
    #     J_m[len(pvpq) + pq_lookup[t[f_in_pvpq & t_in_pq]], pvpq_lookup[f[f_in_pvpq & t_in_pq]]] = np.where(control_mode_v, SMALL_NUMBER, S_Fik.real)  # control mode V or Q
    # if np.any(t_in_pq & t_in_pvpq):
    #     J_m[len(pvpq) + pq_lookup[t[t_in_pq]], pvpq_lookup[t[t_in_pvpq]]] = np.where(control_mode_v, SMALL_NUMBER, -S_Fik.real) # control mode V or Q

    rows = np.r_[rows, len(pvpq) + pq_lookup[f[f_in_pq]], len(pvpq) + pq_lookup[f[f_in_pq & t_in_pvpq]], len(pvpq) + pq_lookup[t[f_in_pvpq & t_in_pq]], len(pvpq) + pq_lookup[t[t_in_pq]]]
    cols = np.r_[cols, pvpq_lookup[f[f_in_pvpq]], pvpq_lookup[t[f_in_pq & t_in_pvpq]], pvpq_lookup[f[f_in_pvpq & t_in_pq]], pvpq_lookup[t[t_in_pvpq]]]
    data = np.r_[data, P_Fik[f_in_pq], -P_Fik[f_in_pq & t_in_pvpq],
                 np.where(control_v[f_in_pvpq & t_in_pq], zero[f_in_pvpq & t_in_pq], P_Fik[f_in_pvpq & t_in_pq]),
                 np.where(control_v[t_in_pq], zero[t_in_pq], -P_Fik[t_in_pq])]
    # data = np.r_[data, P_Fik[f_in_pq], -P_Fik[f_in_pq & t_in_pvpq],
    #              np.where(control_v[f_in_pvpq & t_in_pq], zero[f_in_pvpq & t_in_pq], P_Fki[f_in_pvpq & t_in_pq]),
    #              np.where(control_v[t_in_pq], zero[t_in_pq], -P_Fki[t_in_pq])]



    # # J_C_Q_u = np.zeros(shape=(len(pq), len(pq)), dtype=np.float64)
    # # J_C_Q_u = np.zeros(shape=(len(pq)+ len(x_control), len(pq)+ len(x_control)), dtype=np.float64)
    # # if np.any(f_in_pq):
    #     # J_C_Q_u[pq_lookup[f[f_in_pq]], pq_lookup[f[f_in_pq]]] = (2 * S_Fii.imag + S_Fik.imag) / Vmf
    #     # J_C_Q_u[pq_lookup[f[f_in_pq]], pq_lookup[t[f_in_pq]]] = S_Fik.imag/Vmt
    #     #
    #     # J_C_Q_u[pq_lookup[t[f_in_pq]], pq_lookup[f[f_in_pq]]] = 1
    #     # J_C_Q_u[pq_lookup[t[f_in_pq]], pq_lookup[t[f_in_pq]]] = 0
    # if np.any(f_in_pq):
    #     J_m[len(pvpq)+pq_lookup[f[f_in_pq]], len(pvpq)+pq_lookup[f[f_in_pq]]] = (2 * S_Fii.imag + S_Fik.imag) / Vmf
    # if np.any(f_in_pq & t_in_pq):
    #     J_m[len(pvpq)+pq_lookup[f[f_in_pq & t_in_pq]], len(pvpq)+pq_lookup[t[f_in_pq & t_in_pq]]] = S_Fik.imag / Vmt
    #     J_m[len(pvpq)+pq_lookup[t[f_in_pq & t_in_pq]], len(pvpq)+pq_lookup[f[f_in_pq & t_in_pq]]] = np.where(control_mode_v, 1, (2 * S_Fii.imag + S_Fik.imag) / Vmf)  # control mode V or Q
    # if np.any(t_in_pq):
    #     J_m[len(pvpq)+pq_lookup[t[t_in_pq]], len(pvpq)+pq_lookup[t[t_in_pq]]] = np.where(control_mode_v, SMALL_NUMBER, S_Fik.imag / Vmt)  # control mode V or Q
    #     #
    #     # J_C_Q_u[pq_lookup[t[f_in_pq]]+ len(x_control), pq_lookup[f[f_in_pq]]] = 1
    #     # J_C_Q_u[pq_lookup[t[f_in_pq]]+ len(x_control), pq_lookup[t[f_in_pq]]+ len(x_control)] = 0

    rows = np.r_[rows, len(pvpq)+pq_lookup[f[f_in_pq]], len(pvpq)+pq_lookup[f[f_in_pq & t_in_pq]], len(pvpq)+pq_lookup[t[f_in_pq & t_in_pq]], len(pvpq)+pq_lookup[t[t_in_pq]]]
    cols = np.r_[cols, len(pvpq)+pq_lookup[f[f_in_pq]], len(pvpq)+pq_lookup[t[f_in_pq & t_in_pq]], len(pvpq)+pq_lookup[f[f_in_pq & t_in_pq]], len(pvpq)+pq_lookup[t[t_in_pq]]]
    data = np.r_[data, ((2 * Q_Fii + Q_Fik) / Vmf)[f_in_pq], (Q_Fik / Vmt)[f_in_pq & t_in_pq],
                 np.where(control_v[f_in_pq & t_in_pq], one[f_in_pq & t_in_pq], ((2 * Q_Fii + Q_Fik) / Vmf)[f_in_pq & t_in_pq]),
                 np.where(control_v[t_in_pq], zero[t_in_pq], (Q_Fik / Vmt)[t_in_pq])]
    # data = np.r_[data, ((2 * Q_Fii + Q_Fik) / Vmf)[f_in_pq], (Q_Fik / Vmt)[f_in_pq & t_in_pq],
    #              np.where(control_v[f_in_pq & t_in_pq], one[f_in_pq & t_in_pq], ((2 * Q_Fkk + Q_Fki) / Vmf)[f_in_pq & t_in_pq]),
    #              np.where(control_v[t_in_pq], zero[t_in_pq], (Q_Fkk / Vmt)[t_in_pq])]

    # for vsc ac slack buses:
    # rows = np.r_[rows, ac_slack]
    # cols = np.r_[cols, ac_slack]
    # data = np.r_[data, np.ones_like(ac_slack, dtype=np.float64)]

    # J_C_P_c = np.zeros(shape=(len(pvpq), nsvc + ntcsc), dtype=np.float64)
    # J_C_Q_c = np.zeros(shape=(len(pq), nsvc + ntcsc + 2 * nssc), dtype=np.float64)
    # J_C_C_d = np.zeros(shape=(nsvc + ntcsc + 2 * nssc, len(pvpq)), dtype=np.float64)
    # J_C_C_u = np.zeros(shape=(nsvc + ntcsc + 2 * nssc, len(pq)), dtype=np.float64)
    # J_C_C_c = np.zeros(shape=(nsvc + ntcsc + 2 *nssc, nsvc + ntcsc + 2 *nssc), dtype=np.float64)



    # if np.any(tcsc_controllable) or np.any(svc_controllable):  # todo
    #     relevant = np.r_[np.arange(nsvc), nsvc + np.arange(ntcsc)[tcsc_controllable]]
    #     J_m = np.vstack([np.hstack([J_C_P_d, J_C_P_u, J_C_P_c[:, relevant]]),
    #                      np.hstack([J_C_Q_d, J_C_Q_u, J_C_Q_c[:, relevant]]),
    #                      np.hstack([J_C_C_d[relevant, :], J_C_C_u[relevant, :],
    #                                 J_C_C_c[:, relevant][relevant, :]])])
    # else:
    #     J_m = np.vstack([np.hstack([J_C_P_d, J_C_P_u]),
    #                      np.hstack([J_C_Q_d, J_C_Q_u])])

    #J_m = csr_matrix(J_m)
    J_m = csr_matrix((data, (rows, cols)), shape=J.shape, dtype=np.float64)

    return J_m


@np.errstate(all="raise")
def create_J_modification_hvdc(J, V_dc, Ybus_hvdc, Ybus_vsc_dc, vsc_g_pu, vsc_gl_pu, dc_p, dc_p_lookup,
                               vsc_dc_fb, vsc_dc_tb, vsc_dc_slack, vsc_dc_mode_p):
    # Calculate the first matrix for all elements
    J_all = (Ybus_hvdc - Ybus_vsc_dc).multiply(V_dc)
    # Calculate the second matrix for diagonal elements only
    J_diag = diags(V_dc).multiply((Ybus_hvdc - Ybus_vsc_dc).dot(V_dc))

    # Calculate DC modification VSC Jacobian
    # here we do not use Ybus because it has the y values already summed up for the diagonal entries,
    # and creating the sparse matrix for J also sums values that share the same indices.
    # So effectively we achieve the same effect here as if we had the Ybus matrix,
    # except that we operate at the level of individual VSC still.
    # Only relevant when more than one VSC connected to the same bus.
    Vmf = V_dc[vsc_dc_fb]
    Vmt = V_dc[vsc_dc_tb]
    P_Fii = Vmf * (vsc_g_pu + vsc_gl_pu) * Vmf
    P_Fik = -Vmf * vsc_g_pu * Vmt
    P_Fkk = Vmt * vsc_g_pu * Vmt
    P_Fki = -Vmt * vsc_g_pu * Vmf

    zero = SMALL_NUMBER * np.ones(shape=P_Fii.shape, dtype=np.float64)
    one = np.ones(shape=P_Fii.shape, dtype=np.float64)

    rows = np.r_[dc_p_lookup[vsc_dc_fb], dc_p_lookup[vsc_dc_fb], dc_p_lookup[vsc_dc_tb], dc_p_lookup[vsc_dc_tb]]
    cols = np.r_[dc_p_lookup[vsc_dc_fb], dc_p_lookup[vsc_dc_tb], dc_p_lookup[vsc_dc_fb], dc_p_lookup[vsc_dc_tb]]
    # data = np.r_[2 * P_Fii + P_Fik, P_Fik, np.where(vsc_dc_slack, one, P_Fki), np.where(vsc_dc_slack, zero, 2 * P_Fkk + P_Fki)]
    # data = np.r_[2 * P_Fii + P_Fik, P_Fik, np.where(vsc_dc_slack, one, 2 * P_Fii + P_Fik), np.where(vsc_dc_slack, zero, P_Fik)]
    # it depends on which power must be controlled, at internal node (AC slack mode) or at external node (DC P mode):
    data = np.r_[(2 * P_Fii + P_Fik) / Vmf, P_Fik / Vmt,
                 np.where(vsc_dc_slack, one, np.where(vsc_dc_mode_p, (2 * P_Fii + P_Fik) / Vmf, P_Fki / Vmf)),
                 np.where(vsc_dc_slack, zero, np.where(vsc_dc_mode_p, P_Fik / Vmt, (2 * P_Fkk + P_Fki) / Vmt))]
    # data = np.r_[P_Fik / V_dc[vsc_dc_fb], P_Fik / V_dc[vsc_dc_tb], np.where(vsc_dc_slack, one, P_Fki / V_dc[vsc_dc_fb]), np.where(vsc_dc_slack, zero, P_Fki / V_dc[vsc_dc_tb])]
    # data = np.r_[-Q_Fik[f_in_pvpq], Q_Fik[f_in_pvpq & t_in_pvpq], np.where(control_delta[f_in_pvpq & t_in_pvpq], one[f_in_pvpq & t_in_pvpq], Q_Fki[f_in_pvpq & t_in_pvpq]), np.where(control_delta[t_in_pvpq], zero[t_in_pvpq], -Q_Fki[t_in_pvpq])]

    J_m_vsc = csr_matrix((data, (rows, cols)), shape=J_all.shape, dtype=np.float64)

    # Combine them to form the Jacobian for DC grid including the VSC elements
    J_combined = J_all + J_diag + J_m_vsc

    # Map the J for the DC system to the overall J
    num_p = len(dc_p)
    offset = J.shape[0] - num_p
    # Create an initial zero sparse matrix for J_m
    J_m = lil_matrix(J.shape)
    # Create the modification J for the DC system to be added to the overall J
    J_m[offset:offset+num_p, offset:offset+num_p] = J_combined[dc_p, :][:, dc_p]

    return J_m.tocsr()




# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import logging
import os
import time
from typing import Union, List, Type, Dict
from pandapower.converter.ucte.ucte_converter import UCTE2pandapower
from pandapower.converter.ucte.ucte_parser import UCTEParser

logger = logging.getLogger('ucte.from_ucte')


def from_ucte_dict(ucte_parser: UCTEParser, slack_as_gen: bool = True):
    """Creates a pandapower net from an UCTE data structure.

    Parameters
    ----------
    ucte_parser : UCTEParser
        The UCTEParser with parsed UCTE data.

    slack_as_gen : bool
        decides whether slack elements are converted as gen or ext_grid elements.

    Returns
    -------
    pandapowerNet
        net
    """

    ucte_converter = UCTE2pandapower(slack_as_gen=slack_as_gen)
    net = ucte_converter.convert(ucte_parser.get_data())

    return net


def from_ucte(ucte_file: str, slack_as_gen: bool = True):
    """Converts net data stored as an UCTE file to a pandapower net.

    Parameters
    ----------
    ucte_file : str
        path to the ucte file which includes all the data of the grid (EHV or HV or both)

    slack_as_gen : bool
        decides whether slack elements are converted as gen or ext_grid elements.

    Returns
    -------
    pandapowerNet
        net

    Example
    -------
    >>> import os
    >>> import pandapower as pp
    >>> ucte_file = os.path.join(pp.pp_dir, "test", "converter", "testfiles", "test_ucte_DK.uct")
    >>> net = pp.converter.from_ucte(ucte_file)
    """

    # Note:
    # the converter functionality from_ucte() and internal functions are structured similar to
    # the cim converter

    time_start_parsing = time.time()

    ucte_parser = UCTEParser(ucte_file)
    ucte_parser.parse_file()

    time_start_converting = time.time()

    pp_net = from_ucte_dict(ucte_parser, slack_as_gen=slack_as_gen)

    time_end_converting = time.time()

    logger.info("Needed time for parsing from ucte: %s" % (time_start_converting - time_start_parsing))
    logger.info("Needed time for converting from ucte: %s" % (time_end_converting - time_start_converting))
    logger.info("Total Time (from_ucte()): %s" % (time_end_converting - time_start_parsing))

    return pp_net


if __name__ == "__main__":
    import os
    import pandapower as pp

    ### loading the line test as example
    ucte_file = os.path.join(pp.pp_dir, "test", "converter", "testfiles", "test_ucte_DK.uct")
    net = pp.converter.from_ucte(ucte_file)



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import logging
import math
import time
from typing import Dict, Union

import numpy as np
import pandapower as pp
import pandapower.auxiliary
import pandas as pd


class UCTE2pandapower:
    def __init__(self, slack_as_gen: bool = True):
        """
        Convert UCTE data to pandapower.
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.u_d = dict()
        self.net = self._create_empty_network()
        self.net.bus["node_name"] = ""
        self.slack_as_gen = slack_as_gen

    @staticmethod
    def _create_empty_network():
        net: pandapower.auxiliary.pandapowerNet = pp.create_empty_network()
        new_columns = {
            "trafo": {
                "tap2_min": int,
                "tap2_max": int,
                "tap2_neutral": int,
                "tap2_pos": int,
                "tap2_step_percent": float,
                "tap2_step_degree": float,
                "tap2_side": str,
                "tap2_changer_type": str,
                "amica_name": str,
            },
            "line": {"amica_name": str},
            "bus": {"ucte_country": str},
        }
        for pp_element in new_columns.keys():
            for col, dtype in new_columns[pp_element].items():
                net[pp_element][col] = pd.Series(dtype=dtype)
        return net

    def convert(self, ucte_dict: Dict) -> pandapower.auxiliary.pandapowerNet:
        self.logger.info("Converting UCTE data to a pandapower network.")
        time_start = time.time()
        # create a temporary copy from the origin input data
        self.u_d = dict()
        for ucte_element, items in ucte_dict.items():
            self.u_d[ucte_element] = items.copy()
        # first reset the index to get indices for pandapower
        for ucte_element in self.u_d.keys():
            if ucte_element == "R":
                continue
            self.u_d[ucte_element] = self.u_d[ucte_element].reset_index(level=0)
            self.u_d[ucte_element] = self.u_d[ucte_element].rename(
                columns={"index": "id"}
            )
        # now replace the node1 and node2 columns with the node index at lines, transformers, ...
        merge_nodes = self.u_d["N"][["id", "node"]]
        self.u_d["L"] = pd.merge(
            self.u_d["L"],
            merge_nodes.rename(columns={"node": "node1", "id": "from_bus"}),
            how="left",
            on="node1",
        )
        self.u_d["L"] = pd.merge(
            self.u_d["L"],
            merge_nodes.rename(columns={"node": "node2", "id": "to_bus"}),
            how="left",
            on="node2",
        )
        self.u_d["L"] = self.u_d["L"].drop(columns=["node1", "node2"])
        for one_asset in ["T", "R", "TT"]:
            self.u_d[one_asset] = pd.merge(
                self.u_d[one_asset],
                merge_nodes.rename(columns={"node": "node1", "id": "hv_bus"}),
                how="left",
                on="node1",
            )
            self.u_d[one_asset] = pd.merge(
                self.u_d[one_asset],
                merge_nodes.rename(columns={"node": "node2", "id": "lv_bus"}),
                how="left",
                on="node2",
            )
            self.u_d[one_asset] = self.u_d[one_asset].drop(columns=["node1", "node2"])

        # prepare the element tables
        self._convert_nodes()
        self._convert_loads()
        self._convert_gens()
        self._convert_lines()
        self._convert_impedances()
        self._convert_switches()
        self._convert_trafos()

        # copy data to the element tables of self.net
        self.net = self.set_pp_col_types(self.net)

        # currently, net.bus.name contains the UCTE node name ("Node"), while
        # net.bus.node_name contains the original node name "Node Name". This is changed now:
        cols_in_order = list(pd.Series(self.net.bus.columns).replace("node_name", "ucte_name"))
        self.net.bus = self.net.bus.rename(columns={"name": "ucte_name", "node_name": "name"})[
            cols_in_order]

        self.logger.info(
            "Finished converting the input data to pandapower in %ss."
            % (time.time() - time_start)
        )
        return self.net

    def _copy_to_pp(self, pp_type: str, input_df: pd.DataFrame):
        self.logger.debug(
            "Copy %s datasets to pandapower network with type %s"
            % (input_df.index.size, pp_type)
        )
        if pp_type not in self.net.keys():
            self.logger.warning(
                "Missing pandapower type %s in the pandapower network!" % pp_type
            )
            return
        self.net[pp_type] = pd.concat(
            [
                self.net[pp_type],
                input_df[
                    list(set(self.net[pp_type].columns).intersection(input_df.columns))
                ],
            ],
            ignore_index=True,
            sort=False,
        )

    def _convert_nodes(self):
        self.logger.info("Converting the nodes.")
        nodes = self.u_d[
            "N"
        ]  # Note: Do not use a copy, the columns 'volt_str' and 'node_name' are needed later
        nodes["volt_str"] = nodes["node"].str[6:7]
        volt_map = {
            "0": 750,
            "1": 380,
            "2": 220,
            "3": 150,
            "4": 120,
            "5": 110,
            "6": 70,
            "7": 27,
            "8": 330,
            "9": 500,
            "A": 26,
            "B": 25,
            "C": 24,
            "D": 23,
            "E": 22,
            "F": 21,
            "G": 20,
            "H": 19,
            "I": 18,
            "J": 17,
            "K": 15.7,
            "L": 15,
            "M": 13.7,
            "N": 13,
            "O": 12,
            "P": 11,
            "Q": 9.8,
            "R": 9,
            "S": 8,
            "T": 7,
            "U": 6,
            "V": 5,
            "W": 4,
            "X": 3,
            "Y": 2,
            "Z": 1,
        }
        nodes["vn_kv"] = nodes["volt_str"].map(volt_map)
        nodes["node2"] = nodes["node"].str[:6]
        nodes["grid_area_id"] = nodes["node"].str[:2]
        # drop all voltages at non pu nodes
        nodes.loc[
            (nodes["node_type"] != 2) & (nodes["node_type"] != 3), "voltage"
        ] = np.nan
        nodes = nodes.rename(columns={"node": "name"})
        nodes["in_service"] = True
        self._copy_to_pp("bus", nodes)
        self.logger.info("Finished converting the nodes.")

    def _convert_loads(self):
        self.logger.info("Converting the loads.")
        # select the loads from the nodes and drop not given values
        loads = self.u_d["N"].dropna(subset=["p_load", "q_load"])
        # select all with p != 0 or q != 0
        loads = loads.loc[(loads["p_load"] != 0) | (loads["q_load"] != 0)]
        if not len(loads):
            self.logger.info("Finished converting the loads (no loads existing).")
            return  # Acceleration
        loads = loads.rename(
            columns={"id": "bus", "node": "name", "p_load": "p_mw", "q_load": "q_mvar"}
        )
        # get a new index
        loads = loads.reset_index(level=0, drop=True)
        loads["scaling"] = 1
        loads["in_service"] = True
        loads["const_z_percent"] = 0
        loads["const_i_percent"] = 0
        self._copy_to_pp("load", loads)
        self.logger.info("Finished converting the loads.")

    def _convert_gens(self):
        self.logger.info("Converting the generators.")
        # select the gens from the nodes and drop not given values
        gens = self.u_d["N"].dropna(subset=["p_gen", "q_gen"])
        # select all with p != 0 or q != 0 or voltage != 0
        gens = gens.loc[
            (gens["p_gen"] != 0) | (gens["q_gen"] != 0) | (gens["voltage"] > 0)
        ]
        # change the signing
        gens["p_gen"] = gens["p_gen"] * -1
        gens["q_gen"] = gens["q_gen"] * -1
        gens["min_p_gen"] = gens["min_p_gen"] * -1
        gens["max_p_gen"] = gens["max_p_gen"] * -1
        gens["min_q_gen"] = gens["min_q_gen"] * -1
        gens["max_q_gen"] = gens["max_q_gen"] * -1
        # drop all voltages at non pu nodes
        gens.loc[
            (gens["node_type"] != 2) & (gens["node_type"] != 3), "voltage"
        ] = np.nan
        gens["vm_pu"] = gens["voltage"] / gens["vn_kv"]
        gens = gens.rename(
            columns={
                "id": "bus",
                "node": "name",
                "p_gen": "p_mw",
                "q_gen": "q_mvar",
                "min_p_gen": "min_p_mw",
                "max_p_gen": "max_p_mw",
                "min_q_gen": "min_q_mvar",
                "max_q_gen": "max_q_mvar",
            }
        )
        # get a new index
        gens = gens.reset_index(level=0, drop=True)
        gens["scaling"] = 1
        gens["va_degree"] = 0
        gens["slack_weight"] = 1
        gens["slack"] = False
        gens["current_source"] = True
        gens["in_service"] = True
        self._copy_to_pp("ext_grid" if not self.slack_as_gen else "gen",
                         gens.loc[gens["node_type"] == 3])
        if self.slack_as_gen:
            self.net.gen["slack"] = True
        self._copy_to_pp("gen", gens.loc[gens["node_type"] == 2])
        self._copy_to_pp(
            "sgen", gens.loc[(gens["node_type"] == 0) | (gens["node_type"] == 1)]
        )
        self.logger.info("Finished converting the generators.")

    def _convert_lines(self):
        self.logger.info("Converting the lines.")
        # get the lines
        # status 9 & 1 stands for equivalent line that can be interpreted as impedance
        # status 7 & 2 stands for busbar coupler that can be interpreted as switches
        lines = self.u_d["L"].loc[self.u_d["L"].status.isin([9, 1, 7, 2]) == False, :]
        # definition of busbar coupler is if r, x, b are all zero, but some statuses are wrong, check for those
        lines = lines.drop(
            lines.loc[(lines.r == 0) & (lines.x == 0) & (lines.b == 0)].index
        )
        # also drop lines with x < 0 as they will be modeled as impedances
        lines = lines.drop(lines.loc[lines.x < 0].index)
        if not len(lines):
            self.logger.info("Finished converting the lines (no lines existing).")
            return  # Acceleration
        # lines = self.u_d['L']
        # create the in_service column from the UCTE status
        in_service_map = dict({0: True, 1: True, 2: True, 7: False, 8: False, 9: False})
        lines["in_service"] = lines["status"].map(in_service_map)
        # i in A to i in kA
        lines["max_i_ka"] = lines["i"] / 1e3
        lines["max_i_ka"] = lines["max_i_ka"].fillna(9999)
        lines["c_nf_per_km"] = 1e3 * lines["b"] / (2 * np.pi * 50)
        lines["g_us_per_km"] = 0
        lines["df"] = 1
        lines["parallel"] = 1
        lines["length_km"] = 1
        self._fill_empty_names(lines)
        self._fill_amica_names(lines, ":line")
        lines.loc[lines.x == 0, "x"] = 0.01
        # rename the columns to the pandapower schema
        lines = lines.rename(
            columns={"r": "r_ohm_per_km", "x": "x_ohm_per_km", "name": "name"}
        )
        self._copy_to_pp("line", lines)
        self.logger.info("Finished converting the lines.")

    def _convert_impedances(self):
        self.logger.info("Converting the impedances.")
        # get the impedances
        # status 9 & 1 stands for equivalent line that can be interpreted as impedance
        status_9_1 = self.u_d["L"].status.isin([9, 1])
        # also all lines with negative x convert to impedances
        negative_x = self.u_d["L"].x < 0
        impedances = self.u_d["L"].loc[status_9_1 | negative_x, :]
        self._set_column_to_type(impedances, "from_bus", int)
        impedances = pd.merge(
            impedances,
            self.u_d["N"][["vn_kv"]],
            how="left",
            left_on="from_bus",
            right_index=True,
        )

        trafos_to_impedances = self._get_trafos_modelled_as_impedances()
        impedances = pd.concat([impedances, trafos_to_impedances])

        # create the in_service column from the UCTE status
        in_service_map = dict({0: True, 1: True, 2: True, 7: False, 8: False, 9: False})
        impedances["in_service"] = impedances["status"].map(in_service_map)
        # Convert ohm/km to per unit (pu)
        impedances["sn_mva"] = 10000  # same as PowerFactory
        impedances["z_ohm"] = impedances["vn_kv"] ** 2 / impedances["sn_mva"]
        impedances["rft_pu"] = impedances["r"] / impedances["z_ohm"]
        impedances["rtf_pu"] = impedances["r"] / impedances["z_ohm"]
        impedances["xft_pu"] = impedances["x"] / impedances["z_ohm"]
        impedances["xtf_pu"] = impedances["x"] / impedances["z_ohm"]
        impedances["gf_pu"] = impedances["g"] / impedances["z_ohm"]
        impedances["gt_pu"] = impedances["g"] / impedances["z_ohm"]
        impedances["bf_pu"] = impedances["b"] / impedances["z_ohm"]
        impedances["bt_pu"] = impedances["b"] / impedances["z_ohm"]
        self._fill_empty_names(impedances)
        self._copy_to_pp("impedance", impedances)
        self.logger.info("Finished converting the impedances.")

    def _get_trafos_modelled_as_impedances(self):
        ### get transformers that will be transformed to impedances and append them ###
        trafos = pd.merge(
            self.u_d["T"],
            self.u_d["R"],
            how="left",
            on=["hv_bus", "lv_bus", "order_code"],
        )

        # check for trafos connecting same voltage levels
        trafos_to_impedances = trafos.loc[
            trafos.loc[:, "0_x"].map(lambda s: s[6])
            == trafos.loc[:, "0_x"].map(lambda s: s[15])
        ]

        trafos_to_impedances = trafos_to_impedances.loc[
            trafos_to_impedances.phase_reg_delta_u.isnull()
        ]
        trafos_to_impedances = trafos_to_impedances.loc[
            trafos_to_impedances.angle_reg_theta.isnull()
        ]
        # calculate iron losses in kW
        trafos_to_impedances["pfe_kw"] = (
            trafos_to_impedances.g * trafos_to_impedances.voltage1**2 / 1e3
        )
        # calculate open loop losses in percent of rated current
        trafos_to_impedances["i0_percent"] = (
            (
                (
                    (trafos_to_impedances.b * 1e-6 * trafos_to_impedances.voltage1**2)
                    ** 2
                    + (
                        trafos_to_impedances.g
                        * 1e-6
                        * trafos_to_impedances.voltage1**2
                    )
                    ** 2
                )
                ** 0.5
            )
            * 100
            / trafos_to_impedances.s
        )
        trafos_to_impedances = trafos_to_impedances.loc[
            (trafos_to_impedances.pfe_kw == 0) & (trafos_to_impedances.i0_percent == 0)
        ]
        # rename the columns to the pandapower schema, as voltages are the same we can take voltage1 as vn_kv
        trafos_to_impedances = trafos_to_impedances.rename(
            columns={
                "hv_bus": "from_bus",
                "lv_bus": "to_bus",
                "voltage1": "vn_kv",
                "0_x": 0,
            }
        )
        return trafos_to_impedances

    def _convert_switches(self):
        self.logger.info("Converting the switches.")
        # get the switches
        # status 7 & 2 stands for busbar coupler that can be interpreted as switch
        switches_by_status = self.u_d["L"].status.isin([7, 2])
        # switches are defined by r, x and b equal 0, but some still have the wrong status (or r, x and b not zero)
        lines_rxb_zero = (
            (self.u_d["L"].r == 0) & (self.u_d["L"].x == 0) & (self.u_d["L"].b == 0)
        )
        switches = self.u_d["L"].loc[lines_rxb_zero | switches_by_status, :]

        # create the in_service column from the UCTE status
        in_service_map = dict({0: True, 1: True, 2: True, 7: False, 8: False, 9: False})
        switches["closed"] = switches["status"].map(in_service_map)
        self._set_column_to_type(switches, "from_bus", int)
        switches["type"] = "LS"
        switches["et"] = "b"
        switches["z_ohm"] = 0
        self._fill_empty_names(switches)
        switches = switches.rename(columns={"from_bus": "bus", "to_bus": "element"})
        self._copy_to_pp("switch", switches)
        self.logger.info("Finished converting the switches.")

    def _convert_trafos(self):
        self.logger.info("Converting the transformers.")
        trafos = pd.merge(
            self.u_d["T"],
            self.u_d["R"],
            how="left",
            on=["hv_bus", "lv_bus", "order_code"],
        )
        if not len(trafos):
            self.logger.info("Finished converting the transformers (no transformers existing).")
            return
        # create the in_service column from the UCTE status
        status_map = dict({0: True, 1: True, 8: False, 9: False})
        trafos["in_service"] = trafos["status"].map(status_map)
        # use same value as in powerfactory for replacing s equals zero values
        trafos.loc[trafos.s == 0, "s"] = 1001
        # calculate the derating factor
        trafos["df"] = trafos["voltage1"] * (trafos["i"] / 1e3) * 3**0.5 / trafos["s"]
        # calculate the relative short-circuit voltage
        trafos["vk_percent"] = (
            np.sign(trafos.x)
            * (abs(trafos.r) ** 2 + abs(trafos.x) ** 2) ** 0.5
            * (trafos.s * 1e3)
            / (10.0 * trafos.voltage1**2)
        )
        # calculate vkr_percent
        trafos["vkr_percent"] = trafos.r * trafos.s * 100 / trafos.voltage1**2
        # calculate iron losses in kW
        trafos["pfe_kw"] = trafos.g * trafos.voltage1**2 / 1e3
        # calculate open loop losses in percent of rated current
        trafos["i0_percent"] = (
            (
                (
                    (trafos.b * 1e-6 * trafos.voltage1**2) ** 2
                    + (trafos.g * 1e-6 * trafos.voltage1**2) ** 2
                )
                ** 0.5
            )
            * 100
            / trafos.s
        )

        # phase and angle regulation have to be split up into 5 cases:
        # only phase regulated -> pr
        # only angle regulated symmetrical model -> ars
        # only angle regulated asymmetrical model -> ara
        # phase and angle regulated symmetrical model -> pars
        # phase and angle regulated asymmetrical model -> para
        # set values for only phase regulated transformers (pr)
        has_phase_values = (
            (~trafos.phase_reg_delta_u.isnull())
            & (~trafos.phase_reg_n.isnull())
            & (~trafos.phase_reg_n2.isnull())
        )
        has_missing_angle_values = (
            trafos.angle_reg_delta_u.isnull()
            | trafos.angle_reg_theta.isnull()
            | trafos.angle_reg_n.isnull()
            | trafos.angle_reg_n2.isnull()
        )
        pr = trafos.loc[has_phase_values & has_missing_angle_values].index

        trafos.loc[pr, "tap_min"] = -trafos["phase_reg_n"]
        trafos.loc[pr, "tap_max"] = trafos["phase_reg_n"]
        trafos.loc[pr, "tap_pos"] = trafos["phase_reg_n2"]
        trafos.loc[pr, "tap_step_percent"] = trafos.loc[pr, "phase_reg_delta_u"].abs()
        trafos.loc[pr, "tap_changer_type"] = "Ratio"

        # set values for only angle regulated transformers symmetrical and asymmetrical
        has_missing_phase_values = (
            trafos.phase_reg_delta_u.isnull()
            & trafos.phase_reg_n.isnull()
            & trafos.phase_reg_n2.isnull()
        )
        has_angle_values = (
            (~trafos.angle_reg_delta_u.isnull())
            & (~trafos.angle_reg_theta.isnull())
            & (~trafos.angle_reg_n.isnull())
            & (~trafos.angle_reg_n2.isnull())
        )
        ar = trafos.loc[has_missing_phase_values & has_angle_values].index

        symm = trafos.angle_reg_type == "SYMM"
        ars = trafos.loc[has_missing_phase_values & has_angle_values & symm].index
        trafos.loc[ars, "tap_min"] = -trafos.loc[ar, "angle_reg_n"]
        trafos.loc[ars, "tap_max"] = trafos.loc[ar, "angle_reg_n"]
        trafos.loc[ars, "tap_pos"] = trafos.loc[ar, "angle_reg_n2"]
        trafos.loc[ars, "tap_step_percent"] = np.nan
        # trafos.loc[ars, 'phase_reg_n'] = trafos.loc[ar, 'angle_reg_n']
        trafos.loc[ars, "tap_changer_type"] = "Ideal"
        trafos.loc[
            ars, "tap_step_degree"
        ] = self._calculate_tap_step_degree_symmetrical(trafos.loc[ars])

        asym = (trafos.angle_reg_type == "ASYM") | (trafos.angle_reg_type == "")
        ara = trafos.loc[has_missing_phase_values & has_angle_values & asym].index
        trafos.loc[ara, "tap2_min"] = -trafos.loc[ar, "angle_reg_n"]
        trafos.loc[ara, "tap2_max"] = trafos.loc[ar, "angle_reg_n"]
        trafos.loc[ara, "tap2_pos"] = trafos.loc[ar, "angle_reg_n2"]
        trafos.loc[ara, "tap2_neutral"] = 0
        trafos.loc[ara, "tap2_step_percent"] = np.nan
        trafos.loc[ara, "tap2_changer_type"] = "Ideal"
        trafos.loc[
            ara, "tap2_step_degree"
        ] = self._calculate_tap_step_degree_asymmetrical(trafos.loc[ara])

        trafos.loc[ara, "tap_min"] = -trafos.loc[ara, "angle_reg_n"]
        trafos.loc[ara, "tap_max"] = trafos.loc[ara, "angle_reg_n"]
        trafos.loc[ara, "tap_pos"] = trafos.loc[ara, "angle_reg_n2"]
        trafos.loc[ara, "tap_changer_type"] = "Ratio"
        trafos.loc[
            ara, "tap_step_percent"
        ] = self._calculate_tap_step_percent_asymmetrical(trafos.loc[ara])

        # get phase and angle regulated transformers symmetrical and asymmetrical
        par = trafos.loc[has_phase_values & has_angle_values].index

        trafos.loc[par, "tap_step_percent"] = trafos.loc[par, "phase_reg_delta_u"].abs()
        trafos.loc[par, "tap_min"] = -trafos.loc[par, "phase_reg_n"]
        trafos.loc[par, "tap_max"] = trafos.loc[par, "phase_reg_n"]
        trafos.loc[par, "tap_pos"] = trafos.loc[par, "phase_reg_n2"]
        trafos.loc[par, "tap_changer_type"] = "Ratio"

        trafos.loc[par, "tap2_min"] = -trafos.loc[par, "angle_reg_n"]
        trafos.loc[par, "tap2_max"] = trafos.loc[par, "angle_reg_n"]
        trafos.loc[par, "tap2_neutral"] = 0
        trafos.loc[par, "tap2_pos"] = trafos.loc[par, "angle_reg_n2"]
        trafos.loc[par, "tap2_step_percent"] = np.nan
        trafos.loc[par, "tap2_changer_type"] = "Ideal"

        pars = trafos.loc[has_phase_values & has_angle_values & symm].index
        trafos.loc[
            pars, "tap2_step_degree"
        ] = self._calculate_tap_step_degree_symmetrical(trafos.loc[pars])

        para = trafos.loc[has_phase_values & has_angle_values & asym].index
        trafos.loc[
            para, "tap2_step_degree"
        ] = self._calculate_tap_step_degree_asymmetrical(trafos.loc[para])
        trafos.loc[para, "tap_step_percent"] = trafos.loc[
            para, "tap_step_percent"
        ] + self._calculate_tap_step_percent_asymmetrical(trafos.loc[para])

        # change signs of tap pos for negative degree or percentage values, since pp only allows positive values
        # trafos.loc[trafos.tap_step_percent < 0, ['tap_pos', 'tap_step_percent']] = trafos.loc[trafos.tap_step_percent < 0, ['tap_pos', 'tap_step_percent']] * -1
        # trafos.loc[trafos.tap_step_degree < 0, ['tap_pos', 'tap_step_degree']] = trafos.loc[trafos.tap_step_degree < 0, ['tap_pos', 'tap_step_degree']] * -1
        # trafos.loc[trafos.tap2_step_degree < 0, ['tap2_pos', 'tap2_step_degree']] = trafos.loc[trafos.tap2_step_degree < 0, ['tap2_pos', 'tap2_step_degree']] * -1

        # set the hv and lv voltage sides to voltage1 and voltage2 (The non-regulated transformer side is currently
        # voltage1, not the hv side!)
        trafos["vn_hv_kv"] = trafos[["voltage1", "voltage2"]].max(axis=1)
        trafos["vn_lv_kv"] = trafos[["voltage1", "voltage2"]].min(axis=1)
        # swap the 'fid_node_start' and 'fid_node_end' if need
        trafos["swap"] = trafos["vn_hv_kv"] != trafos["voltage1"]
        # copy the 'fid_node_start' and 'fid_node_end'
        trafos["hv_bus2"] = trafos["hv_bus"].copy()
        trafos["lv_bus2"] = trafos["lv_bus"].copy()
        trafos.loc[trafos.swap, "hv_bus"] = trafos.loc[trafos.swap, "lv_bus2"]
        trafos.loc[trafos.swap, "lv_bus"] = trafos.loc[trafos.swap, "hv_bus2"]
        # set the tap side, default is lv Correct it for other windings
        trafos["tap_side"] = "lv"
        trafos["tap2_side"] = "lv"
        trafos.loc[trafos.swap, "tap_side"] = "hv"
        trafos.loc[trafos.swap, "tap2_side"] = "hv"
        # now set it to nan for not existing tap changers
        trafos.loc[trafos.phase_reg_n.isnull(), "tap_side"] = None
        trafos["tap_neutral"] = 0
        trafos.loc[trafos.phase_reg_n.isnull(), "tap_neutral"] = np.nan
        trafos["shift_degree"] = 0
        trafos["parallel"] = 1
        self._fill_empty_names(trafos, "0_x")
        self._fill_amica_names(trafos, ":trf", "0_x")
        trafos["tap_changer_type"] = trafos["tap_changer_type"].fillna("Ratio").astype(str)
        trafos["tap2_changer_type"] = trafos["tap2_changer_type"].fillna("Ratio").astype(str)
        # rename the columns to the pandapower schema
        trafos = trafos.rename(columns={"s": "sn_mva"})
        # drop transformers that will be transformed to impedances
        trafos_to_impedances = self._get_trafos_modelled_as_impedances()
        trafos = trafos.drop(trafos_to_impedances.index)

        self._copy_to_pp("trafo", trafos)
        self.logger.info("Finished converting the transformers.")

    def _calculate_tap_step_degree_symmetrical(
        self, input_df: pd.DataFrame
    ) -> pd.Series:
        return input_df["angle_reg_delta_u"].apply(lambda du: 2 * math.atan2(du, 2))

    def _calculate_tap_step_degree_asymmetrical(
        self, input_df: pd.DataFrame
    ) -> pd.Series:
        numerator = input_df["angle_reg_delta_u"] * input_df["angle_reg_theta"].apply(
            lambda t: math.sin(t)
        )
        denominator = 1 + (
            input_df["angle_reg_delta_u"]
            * input_df["angle_reg_theta"].apply(lambda t: math.cos(t))
        )
        return (numerator / denominator).apply(lambda t: math.atan2(t, 1))

    def _calculate_tap_step_percent_asymmetrical(
        self, input_df: pd.DataFrame
    ) -> pd.Series:
        sine_sq = (
            input_df["angle_reg_delta_u"]
            * input_df["angle_reg_theta"].apply(lambda t: math.sin(t))
        ) ** 2
        cosine_sq = (
            input_df["angle_reg_delta_u"]
            * input_df["angle_reg_theta"].apply(lambda t: math.cos(t))
        ) ** 2
        tab_step_percent = (sine_sq + cosine_sq).apply(lambda x: np.sqrt(x)) - 1
        return tab_step_percent

    def _set_column_to_type(self, input_df: pd.DataFrame, column: str, data_type):
        try:
            input_df[column] = input_df[column].astype(data_type)
        except Exception as e:
            self.logger.error(
                "Couldn't set data type %s for column %s!" % (data_type, column)
            )
            self.logger.exception(e)

    def _fill_empty_names(self, input_df: pd.DataFrame, input_column: Union[str, int] = 0):
        """Fills empty names with node1_node2_order-code"""

        def get_name_from_ucte_string(ucte_string: str) -> str:
            return f"{ucte_string[:8].strip()}_{ucte_string[9:17].strip()}_{ucte_string[18]}"

        new_names = input_df.loc[input_df["name"] == "", input_column].map(
            get_name_from_ucte_string
        )
        input_df.loc[input_df["name"] == "", "name"] = new_names

    def _fill_amica_names(
        self, input_df: pd.DataFrame, suffix: str, input_column: Union[str, int] = 0
    ) -> None:
        def get_name_from_ucte_string(ucte_string: str) -> str:
            node1 = ucte_string[:7].replace(" ", "_")
            node2 = ucte_string[9:16].replace(" ", "_")
            order_code = ucte_string[18]
            # for some cases the element name is taken instead of the order code, but not clear when
            # taking always the order code leads to more matches, at least for lines...
            return f"{node1}_{node2}_{order_code}{suffix}"

        amica_names = input_df.loc[:, input_column].map(get_name_from_ucte_string)
        input_df.loc[:, "amica_name"] = amica_names

    def set_pp_col_types(
        self,
        net: Union[pandapower.auxiliary.pandapowerNet, Dict],
        ignore_errors: bool = False,
    ) -> pandapower.auxiliary.pandapowerNet:
        """
        Set the data types for some columns from pandapower assets. This mainly effects bus columns (to int, e.g.
        sgen.bus or line.from_bus) and in_service and other boolean columns (to bool, e.g. line.in_service or gen.slack).
        :param net: The pandapower network to update the data types.
        :param ignore_errors: Ignore problems if set to True (no warnings displayed). Optional, default: False.
        :return: The pandapower network with updated data types.
        """
        time_start = time.time()
        pp_elements = [
            "bus",
            "dcline",
            "ext_grid",
            "gen",
            "impedance",
            "line",
            "load",
            "sgen",
            "shunt",
            "storage",
            "switch",
            "trafo",
            "trafo3w",
            "ward",
            "xward",
        ]
        to_int = ["bus", "element", "to_bus", "from_bus", "hv_bus", "mv_bus", "lv_bus"]
        to_bool = ["in_service", "closed"]
        self.logger.info(
            "Setting the columns data types for buses to int and in_service to bool for the following elements: "
            "%s" % pp_elements
        )
        int_type = int
        bool_type = bool
        for ele in pp_elements:
            self.logger.info("Accessing pandapower element %s." % ele)
            if not hasattr(net, ele):
                if not ignore_errors:
                    self.logger.warning(
                        "Missing the pandapower element %s in the input pandapower network!"
                        % ele
                    )
                continue
            for one_int in to_int:
                if one_int in net[ele].columns:
                    self._set_column_to_type(net[ele], one_int, int_type)
            for one_bool in to_bool:
                if one_bool in net[ele].columns:
                    self._set_column_to_type(net[ele], one_bool, bool_type)
        # some individual things
        if hasattr(net, "sgen"):
            self._set_column_to_type(net["sgen"], "current_source", bool_type)
        if hasattr(net, "gen"):
            self._set_column_to_type(net["gen"], "slack", bool_type)
        if hasattr(net, "shunt"):
            self._set_column_to_type(net["shunt"], "step", int_type)
            self._set_column_to_type(net["shunt"], "max_step", int_type)
        self.logger.info(
            "Finished setting the data types for the pandapower network in %ss."
            % (time.time() - time_start)
        )
        return net



import logging
import time

import pandas as pd

from pandapower.converter.cim import cim_tools
from pandapower.converter.cim.cim2pp import build_pp_net
from pandapower.converter.cim.other_classes import Report, LogLevel, ReportCode

logger = logging.getLogger('cim.cim2pp.converter_classes.equivalentInjectionsCim16')

sc = cim_tools.get_pp_net_special_columns_dict()


class EquivalentInjectionsCim16:
    def __init__(self, cimConverter: build_pp_net.CimConverter):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cimConverter = cimConverter

    def convert_equivalent_injections_cim16(self):
        time_start = time.time()
        self.logger.info("Start converting EquivalentInjections.")
        eqssh_ei = self._prepare_equivalent_injections_cim16()
        # split up to wards and xwards: the wards have no regulation
        eqssh_ei_wards = eqssh_ei.loc[~eqssh_ei.regulationStatus]
        eqssh_ei_xwards = eqssh_ei.loc[eqssh_ei.regulationStatus]
        self.cimConverter.copy_to_pp('ward', eqssh_ei_wards)
        self.cimConverter.copy_to_pp('xward', eqssh_ei_xwards)
        self.logger.info("Created %s wards and %s extended ward elements in %ss." %
                         (eqssh_ei_wards.index.size, eqssh_ei_xwards.index.size, time.time() - time_start))
        self.cimConverter.report_container.add_log(Report(
            level=LogLevel.INFO, code=ReportCode.INFO_CONVERTING,
            message="Created %s wards and %s extended ward elements from EquivalentInjections in %ss." %
                    (eqssh_ei_wards.index.size, eqssh_ei_xwards.index.size, time.time() - time_start)))

    def _prepare_equivalent_injections_cim16(self) -> pd.DataFrame:
        if 'sc' in self.cimConverter.cim.keys():
            equivalent_injection = self.cimConverter.merge_eq_other_profiles(['ssh', 'sc'], 'EquivalentInjection',
                                                                             add_cim_type_column=True)
        else:
            equivalent_injection = self.cimConverter.merge_eq_ssh_profile('EquivalentInjection',
                                                                          add_cim_type_column=True)
        eq_base_voltages = pd.concat([self.cimConverter.cim['eq']['BaseVoltage'][['rdfId', 'nominalVoltage']],
                                      self.cimConverter.cim['eq_bd']['BaseVoltage'][['rdfId', 'nominalVoltage']]],
                                     sort=False)
        eq_base_voltages = eq_base_voltages.drop_duplicates(subset=['rdfId'])
        eq_base_voltages = eq_base_voltages.rename(columns={'rdfId': 'BaseVoltage'})
        equivalent_injection = pd.merge(equivalent_injection, eq_base_voltages, how='left', on='BaseVoltage')
        equivalent_injection = pd.merge(equivalent_injection, self.cimConverter.bus_merge, how='left', on='rdfId')
        # maybe the BaseVoltage is not given, also get the nominalVoltage from the buses
        equivalent_injection = pd.merge(equivalent_injection, self.cimConverter.net.bus[['vn_kv']], how='left',
                                        left_on='index_bus', right_index=True)
        equivalent_injection.nominalVoltage = equivalent_injection.nominalVoltage.fillna(equivalent_injection.vn_kv)
        equivalent_injection['regulationStatus'] = equivalent_injection['regulationStatus'].fillna(False)
        equivalent_injection['vm_pu'] = equivalent_injection.regulationTarget / equivalent_injection.nominalVoltage
        if 'inService' in equivalent_injection.columns:
            equivalent_injection['connected'] = equivalent_injection['connected'] & equivalent_injection['inService']
        equivalent_injection = equivalent_injection.rename(
            columns={'rdfId_Terminal': sc['t'], 'rdfId': sc['o_id'], 'connected': 'in_service', 'index_bus': 'bus',
                     'p': 'ps_mw', 'q': 'qs_mvar'})
        equivalent_injection['pz_mw'] = 0.
        equivalent_injection['qz_mvar'] = 0.
        equivalent_injection['r_ohm'] = 0.
        equivalent_injection['x_ohm'] = .1
        return equivalent_injection



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import gc
import warnings

import numpy as np
import pandas as pd
from packaging.version import Version

from pandapower._version import __version__
from pandapower.auxiliary import ets_to_element_types

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def get_element_index(net, element_type, name, exact_match=True):
    """
    Returns the element(s) identified by a name or regex and its element-table.

    INPUT:
      **net** - pandapower network

      **element_type** - Table to get indices from ("line", "bus", "trafo" etc.)

      **name** - Name of the element to match.

    OPTIONAL:
      **exact_match** (boolean, True) -
          True: Expects exactly one match, raises UserWarning otherwise.
          False: returns all indices containing the name

    OUTPUT:
      **index** - The index (or indices in case of exact_match=False) of matching element(s).
    """
    if exact_match:
        idx = net[element_type][net[element_type]["name"] == name].index
        if len(idx) == 0:
            raise UserWarning(f"There is no {element_type} with name {name}")
        if len(idx) > 1:
            raise UserWarning(f"Duplicate {element_type} names for {name}")
        return idx[0]
    else:
        return net[element_type][net[element_type]["name"].str.contains(name)].index


def get_element_indices(net, element_type, name, exact_match=True):
    """
    Returns a list of element(s) identified by a name or regex and its element-table -> Wrapper
    function of get_element_index()

    INPUT:
      **net** - pandapower network

      **element_type** (str, string iterable) - Element table to get indices from
      ("line", "bus", "trafo" etc.).

      **name** (str) - Name of the element to match.

    OPTIONAL:
      **exact_match** (boolean, True)

          - True: Expects exactly one match, raises UserWarning otherwise.
          - False: returns all indices containing the name

    OUTPUT:
      **index** (list) - List of the indices of matching element(s).

    EXAMPLE:
        >>> from pandapower.networks.create_examples import example_multivoltage
        >>> from pandapower import get_element_indices
        >>> net = example_multivoltage()
        >>> # get indices of only one element type (buses in this example):
        >>> get_element_indices(net, "bus", ["Bus HV%i" % i for i in range(1, 4)])
        [32, 33, 34]
        >>> # get indices of only two element type (first buses, second lines):
        >>> get_element_indices(net, ["bus", "line"], "HV", exact_match=False)
        [Int64Index([32, 33, 34, 35], dtype='int64'), Int64Index([0, 1, 2, 3, 4, 5], dtype='int64')]
        >>> get_element_indices(net, ["bus", "line"], ["Bus HV3", "MV Line6"])
        [34, 11]
    """
    if isinstance(element_type, str) and isinstance(name, str):
        element_type = [element_type]
        name = [name]
    else:
        element_type = element_type if not isinstance(element_type, str) else \
            [element_type] * len(name)
        name = name if not isinstance(name, str) else [name] * len(element_type)
    if len(element_type) != len(name):
        raise ValueError("'element_type' and 'name' must have the same length.")
    idx = []
    for et, nam in zip(element_type, name):
        idx += [get_element_index(net, et, nam, exact_match=exact_match)]
    return idx


def next_bus(net, bus, element_id, et='line', **kwargs):
    """
    Returns the index of the second bus an element is connected to, given a
    first one. E.g. the from_bus given the to_bus of a line.

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    bus : int
        index of bus
    element_id : int
        index of element
    et : str, optional
        which branch element type to consider, by default 'line'

    Returns
    -------
    int
        index of next connected bus
    """
    # todo: what to do with trafo3w?
    if et == 'line' or et == 'l':
        bc = ["from_bus", "to_bus"]
    elif et == 'trafo' or et == 't':
        bc = ["hv_bus", "lv_bus"]
    elif et == "switch" and list(net[et].loc[element_id, ["et"]].values) == ['b']:
        # Raises error if switch is not a bus-bus switch
        bc = ["bus", "element"]
    else:
        raise Exception("unknown element type")
    nb = list(net[et].loc[element_id, bc].values)
    nb.remove(bus)
    return nb[0]


def get_connected_elements(net, element_type, buses, respect_switches=True, respect_in_service=False):
    """
     Returns elements connected to a given buses.

     INPUT:
        **net** (pandapowerNet)

        **element_type** (string, name of the element table)

        **buses** (single integer or iterable of ints)

     OPTIONAL:
        **respect_switches** (boolean, True)

            - True: open switches will be respected
            - False: open switches will be ignored

        **respect_in_service** (boolean, False)

            - True: in_service status of connected lines will be respected
            - False: in_service status will be ignored

     OUTPUT:
        **connected_elements** (set) - Returns connected elements.

    """

    if not hasattr(buses, "__iter__"):
        buses = [buses]

    if element_type in ["line", "l"]:
        element_type = "l"
        element_table = net.line
        connected_elements = set(net.line.index[net.line.from_bus.isin(buses) |
                                                net.line.to_bus.isin(buses)])

    elif element_type in ["dcline"]:
        element_table = net.dcline
        connected_elements = set(net.dcline.index[net.dcline.from_bus.isin(buses) |
                                                  net.dcline.to_bus.isin(buses)])

    elif element_type in ["trafo"]:
        element_type = "t"
        element_table = net.trafo
        connected_elements = set(net["trafo"].index[(net.trafo.hv_bus.isin(buses)) |
                                                    (net.trafo.lv_bus.isin(buses))])
    elif element_type in ["trafo3w", "t3w"]:
        element_type = "t3w"
        element_table = net.trafo3w
        connected_elements = set(net["trafo3w"].index[(net.trafo3w.hv_bus.isin(buses)) |
                                                      (net.trafo3w.mv_bus.isin(buses)) |
                                                      (net.trafo3w.lv_bus.isin(buses))])
    elif element_type == "impedance":
        element_table = net.impedance
        connected_elements = set(net["impedance"].index[(net.impedance.from_bus.isin(buses)) |
                                                        (net.impedance.to_bus.isin(buses))])
    elif element_type == "measurement":
        element_table = net[element_type]
        connected_elements = set(net.measurement.index[(net.measurement.element.isin(buses)) |
                                                       (net.measurement.element_type == "bus")])
    elif element_type in pp_elements(bus=False, branch_elements=False):
        element_table = net[element_type]
        connected_elements = set(element_table.index[(element_table.bus.isin(buses))])
    elif element_type in ['_equiv_trafo3w']:
        # ignore '_equiv_trafo3w'
        return {}
    else:
        raise UserWarning(f"Unknown element type {element_type}!")

    if respect_switches and element_type in ["l", "t", "t3w"]:
        open_switches = get_connected_switches(net, buses, consider=element_type, status="open")
        if open_switches:
            open_and_connected = net.switch.loc[net.switch.index.isin(open_switches) &
                                                net.switch.element.isin(connected_elements)].index
            connected_elements -= set(net.switch.element[open_and_connected])

    if respect_in_service and "in_service" in element_table and not element_table.empty:
        connected_elements -= set(element_table[~element_table.in_service].index)

    return connected_elements


def get_connected_buses(net, buses, consider=("l", "s", "t", "t3", "i"), respect_switches=True,
                        respect_in_service=False):
    """
     Returns buses connected to given buses. The source buses will NOT be returned.

     INPUT:
        **net** (pandapowerNet)

        **buses** (single integer or iterable of ints)

     OPTIONAL:
        **respect_switches** (boolean, True)

            - True: open switches will be respected
            - False: open switches will be ignored

        **respect_in_service** (boolean, False)

            - True: in_service status of connected buses will be respected
            - False: in_service status will be ignored

        **consider** (iterable, ("l", "s", "t", "t3", "i")) - Determines, which types of
        connections will be considered.

            l: lines

            s: switches

            t: trafos

            t3: trafo3ws

            i: impedances

     OUTPUT:
        **cl** (set) - Returns connected buses.

    """
    if not hasattr(buses, "__iter__"):
        buses = [buses]

    cb = set()
    if "l" in consider or 'line' in consider:
        in_service_constr = net.line.in_service if respect_in_service else True
        opened_lines = set(net.switch.loc[(~net.switch.closed) & (
                net.switch.et == "l")].element.unique()) if respect_switches else set()
        connected_fb_lines = set(net.line.index[(
                                                    net.line.from_bus.isin(buses)) & ~net.line.index.isin(
            opened_lines) &
                                                in_service_constr])
        connected_tb_lines = set(net.line.index[(
                                                    net.line.to_bus.isin(buses)) & ~net.line.index.isin(
            opened_lines) & in_service_constr])
        cb |= set(net.line[net.line.index.isin(connected_tb_lines)].from_bus)
        cb |= set(net.line[net.line.index.isin(connected_fb_lines)].to_bus)

    if "s" in consider or 'switch' in consider:
        cs = get_connected_switches(net, buses, consider='b',
                                    status="closed" if respect_switches else "all")
        cb |= set(net.switch[net.switch.index.isin(cs)].element)
        cb |= set(net.switch[net.switch.index.isin(cs)].bus)

    if "t" in consider or 'trafo' in consider:
        in_service_constr = net.trafo.in_service if respect_in_service else True
        opened_trafos = set(net.switch.loc[(~net.switch.closed) & (
                net.switch.et == "t")].element.unique()) if respect_switches else set()
        connected_hvb_trafos = set(net.trafo.index[(
                                                       net.trafo.hv_bus.isin(buses)) & ~net.trafo.index.isin(
            opened_trafos) &
                                                   in_service_constr])
        connected_lvb_trafos = set(net.trafo.index[(
                                                       net.trafo.lv_bus.isin(buses)) & ~net.trafo.index.isin(
            opened_trafos) &
                                                   in_service_constr])
        cb |= set(net.trafo.loc[net.trafo.index.isin(connected_lvb_trafos)].hv_bus.values)
        cb |= set(net.trafo.loc[net.trafo.index.isin(connected_hvb_trafos)].lv_bus.values)

    # Gives the lv mv and hv buses of a 3 winding transformer
    if "t3" in consider or 'trafo3w' in consider:
        in_service_constr3w = net.trafo3w.in_service if respect_in_service else True
        if respect_switches:
            opened_buses_hv = set(net.switch.loc[
                                      ~net.switch.closed & (net.switch.et == "t3") &
                                      net.switch.bus.isin(net.trafo3w.hv_bus)].bus.unique())
            opened_buses_mv = set(net.switch.loc[
                                      ~net.switch.closed & (net.switch.et == "t3") &
                                      net.switch.bus.isin(net.trafo3w.mv_bus)].bus.unique())
            opened_buses_lv = set(net.switch.loc[
                                      ~net.switch.closed & (net.switch.et == "t3") &
                                      net.switch.bus.isin(net.trafo3w.lv_bus)].bus.unique())
        else:
            opened_buses_hv = opened_buses_mv = opened_buses_lv = set()

        hvb_trafos3w = net.trafo3w.index[
            net.trafo3w.hv_bus.isin(buses) & ~net.trafo3w.hv_bus.isin(opened_buses_hv) &
            in_service_constr3w]
        mvb_trafos3w = net.trafo3w.index[
            net.trafo3w.mv_bus.isin(buses) & ~net.trafo3w.mv_bus.isin(opened_buses_mv) &
            in_service_constr3w]
        lvb_trafos3w = net.trafo3w.index[
            net.trafo3w.lv_bus.isin(buses) & ~net.trafo3w.lv_bus.isin(opened_buses_lv) &
            in_service_constr3w]

        cb |= (set(net.trafo3w.loc[hvb_trafos3w].mv_bus) | set(
            net.trafo3w.loc[hvb_trafos3w].lv_bus) - opened_buses_mv - opened_buses_lv)
        cb |= (set(net.trafo3w.loc[mvb_trafos3w].hv_bus) | set(
            net.trafo3w.loc[mvb_trafos3w].lv_bus) - opened_buses_hv - opened_buses_lv)
        cb |= (set(net.trafo3w.loc[lvb_trafos3w].hv_bus) | set(
            net.trafo3w.loc[lvb_trafos3w].mv_bus) - opened_buses_hv - opened_buses_mv)

    if "i" in consider or 'impedance' in consider:
        in_service_constr = net.impedance.in_service if respect_in_service else True
        connected_fb_impedances = set(net.impedance.index[
                                          (net.impedance.from_bus.isin(buses)) & in_service_constr])
        connected_tb_impedances = set(net.impedance.index[
                                          (net.impedance.to_bus.isin(buses)) & in_service_constr])
        cb |= set(net.impedance[net.impedance.index.isin(connected_tb_impedances)].from_bus)
        cb |= set(net.impedance[net.impedance.index.isin(connected_fb_impedances)].to_bus)

    if respect_in_service:
        cb -= set(net.bus[~net.bus.in_service].index)

    return cb - set(buses)


def get_connected_buses_at_element(net, element_index, element_type, respect_in_service=False):
    """
     Returns buses connected to a given branch element. In case of a bus switch, two buses
     will be returned, else one.

     INPUT:
        **net** (pandapowerNet)

        **element_index** (integer)

        **element_type** (string) - Type of the source element:

            l, line: line

            s, switch: switch

            t, trafo: trafo

            t3, trafo3w: trafo3w

            i, impedance: impedance

     OPTIONAL:
        **respect_in_service** (boolean, False)

        True: in_service status of connected buses will be respected

        False: in_service status will be ignored

     OUTPUT:
        **cl** (set) - Returns connected switches.

    """

    cb = set()
    if element_type == 'l' or element_type == 'line':
        cb.add(net.line.from_bus.at[element_index])
        cb.add(net.line.to_bus.at[element_index])
    elif element_type == 's' or element_type == 'switch':
        cb.add(net.switch.bus.at[element_index])
        if net.switch.et.at[element_index] == 'b':
            cb.add(net.switch.element.at[element_index])
    elif element_type == 't' or element_type == 'trafo':
        cb.add(net.trafo.hv_bus.at[element_index])
        cb.add(net.trafo.lv_bus.at[element_index])
    elif element_type == 't3' or element_type == 'trafo3w':
        cb.add(net.trafo3w.hv_bus.at[element_index])
        cb.add(net.trafo3w.mv_bus.at[element_index])
        cb.add(net.trafo3w.lv_bus.at[element_index])
    elif element_type == 'i' or element_type == 'impedance':
        cb.add(net.impedance.from_bus.at[element_index])
        cb.add(net.impedance.to_bus.at[element_index])

    if respect_in_service:
        cb -= set(net.bus[~net.bus.in_service].index)

    return cb


def get_connected_buses_at_switches(net, switches):
    """
    Returns a set of buses connected to given switches.

    INPUT:
        **net** (pandapowerNet)

        **switches** (single integer or iterable of ints)

    OUTPUT:
       **buses** (set) - Returns connected buses
    """
    if not hasattr(switches, "__iter__"):
        switches = [switches]
    switches = np.array(switches)
    buses = set(net.switch.bus.loc[switches])
    bebd = branch_element_bus_dict()
    for element_type, et in zip(["bus", "line", "trafo", "trafo3w"], ["b", "l", "t", "t3"]):
        is_et = (net.switch.et.loc[switches] == et).values
        if not any(is_et):
            continue
        elif et == "b":
            buses |= set(net.switch.element.loc[switches[is_et]])
            continue
        for bus_col in bebd[element_type]:
            buses |= set(net[element_type].loc[net.switch.element.loc[switches[is_et]], bus_col])
    return buses


def get_connected_switches(net, buses, consider=('b', 'l', 't', 't3', 'i'), status="all",
                           include_element_connections=False):
    """
    Returns switches connected to given buses.

    INPUT:
        **net** (pandapowerNet)

        **buses** (single integer or iterable of ints)

    OPTIONAL:
        **consider** (iterable, ("l", "s", "t", "t3)) -  Determines, which types of connections
                                                      will be considered.
                                                      l: lines
                                                      b: bus-bus-switches
                                                      t: transformers
                                                      t3: 3W transformers
                                                      i: impedance

        **status** (string, ("all", "closed", "open")) -  Determines, which switches will
                                                            be considered

        **include_element_connections** (bool, False) - If True, also the other bus of the connected
        element, e.g. the other line ending, is included

    OUTPUT:
       **cl** (set) - Returns connected switches.
    """

    if not hasattr(buses, "__iter__"):
        buses = [buses]

    if status == "closed":
        switch_selection = net.switch.closed
    elif status == "open":
        switch_selection = ~net.switch.closed
    else:
        switch_selection = np.full(len(net.switch), True, dtype=bool)
        if status != "all":
            logger.warning("Unknown switch status \"%s\" selected! "
                           "Selecting all switches by default." % status)

    if include_element_connections:
        bebd = branch_element_bus_dict()

    cs = set()
    for et in consider:
        if et == 'b':
            cs |= set(net['switch'].index[
                          (net['switch']['bus'].isin(buses) |
                           net['switch']['element'].isin(buses)) &
                          (net['switch']['et'] == 'b') & switch_selection])
        else:
            cs |= set(net['switch'].index[(net['switch']['bus'].isin(buses)) &
                                          (net['switch']['et'] == et) & switch_selection])
            if include_element_connections:
                element_type = ets_to_element_types(et)
                sw_idx = net.switch.index[(net.switch.et == et) & switch_selection]
                element_buses = net[element_type].loc[list(net.switch.element.loc[sw_idx]),
                bebd[element_type]]
                isin_df = pd.concat([element_buses[col].isin(buses) for col in element_buses],
                                    axis=1)
                cs |= set(sw_idx[isin_df.any(axis=1)])
    return cs


def get_connected_elements_dict(
        net, buses, respect_switches=True, respect_in_service=False, include_empty_lists=False,
        element_types=None, **kwargs):
    """Returns a dict of lists of connected elements.

    Parameters
    ----------
    net : _type_
        _description_
    buses : iterable of buses
        buses as origin to search for connected elements
    respect_switches : bool, optional
        _description_, by default True
    respect_in_service : bool, optional
        _description_, by default False
    include_empty_lists : bool, optional
        if True, the output doesn't have values of empty lists but may lack of element types as
        keys, by default False
    element_types : iterable of strings, optional
        types elements which are analyzed for connection. If not given, all pandapower element types
        are analyzed. That list of all element types can also be restricted by key word arguments
        "connected_buses", "connected_bus_elements", "connected_branch_elements" and
        "connected_other_elements", by default None

    Returns
    -------
    dict[str,list]
        elements connected to given buses
    """
    if element_types is None:
        element_types = pp_elements(
            bus=kwargs.get("connected_buses", True),
            bus_elements=kwargs.get("connected_bus_elements", True),
            branch_elements=kwargs.get("connected_branch_elements", True),
            other_elements=kwargs.get("connected_other_elements", True),
            cost_tables=False,
            res_elements=False)

    connected = dict()
    for et in element_types:
        if et == "bus":
            conn = get_connected_buses(net, buses, respect_switches=respect_switches,
                                       respect_in_service=respect_in_service)
        elif et == "switch":
            conn = get_connected_switches(net, buses)
        else:
            conn = get_connected_elements(
                net, et, buses, respect_switches=respect_switches,
                respect_in_service=respect_in_service)
        if include_empty_lists or len(conn):
            connected[et] = list(conn)
    return connected


def get_connecting_branches(net, buses1, buses2, branch_elements=None):
    """
    Gets/Drops branches that connects any bus of buses1 with any bus of buses2.
    """
    branch_dict = branch_element_bus_dict(include_switch=True)
    if branch_elements is not None:
        branch_dict = {key: branch_dict[key] for key in branch_elements}
    if "switch" in branch_dict:
        branch_dict["switch"].append("element")

    found = {et: set() for et in branch_dict.keys()}
    for et, bus_types in branch_dict.items():
        for bus1 in bus_types:
            for bus2 in bus_types:
                if bus2 != bus1:
                    idx = net[et].index[net[et][bus1].isin(buses1) & net[et][bus2].isin(buses2)]
                    if et == "switch":
                        idx = idx.intersection(net[et].index[net[et].et == "b"])
                    found[et] |= set(idx)
    return {key: val for key, val in found.items() if len(val)}


def get_gc_objects_dict():
    """
    This function is based on the code in mem_top module
    Summarize object types that are tracket by the garbage collector in the moment.
    Useful to test if there are memoly leaks.
    :return: dictionary with keys corresponding to types and values to the number of objects of the
    type
    """
    objs = gc.get_objects()
    nums_by_types = dict()

    for obj in objs:
        _type = type(obj)
        nums_by_types[_type] = nums_by_types.get(_type, 0) + 1
    return nums_by_types


def false_elm_links(net, element_type, col, target_element_type):
    """
    Returns which indices have links to elements of other element tables which does not exist in the
    net.

    Examples
    --------
    >>> false_elm_links(net, "line", "to_bus", "bus")  # exemplary input 1
    >>> false_elm_links(net, "poly_cost", "element", net["poly_cost"]["et"])  # exemplary input 2
    """
    if isinstance(target_element_type, str):
        return net[element_type][col].index[~net[element_type][col].isin(net[
                                                                             target_element_type].index)]
    else:  # target_element_type is an iterable, e.g. a Series such as net["poly_cost"]["et"]
        df = pd.DataFrame({"element": net[element_type][col].values, "et": target_element_type,
                           "indices": net[element_type][col].index.values})
        df = df.set_index("et")
        false_links = pd.Index([])
        for et in df.index:
            false_links = false_links.union(pd.Index(df.loc[et].indices.loc[
                                                         ~df.loc[et].element.isin(net[et].index)]))
        return false_links


def false_elm_links_loop(net, element_types=None):
    """
    Returns a dict of elements which indices have links to elements of other element tables which
    does not exist in the net.
    This function is an outer loop for get_false_links() applications.
    """
    false_links = dict()
    element_types = element_types if element_types is not None else pp_elements(
        bus=False, cost_tables=True)
    bebd = branch_element_bus_dict(include_switch=True)
    for element_type in element_types:
        if net[element_type].shape[0]:
            fl = pd.Index([])
            # --- define col and target_element_type
            if element_type in bebd.keys():
                for col in bebd[element_type]:
                    fl = fl.union(false_elm_links(net, element_type, col, "bus"))
            elif element_type in {"poly_cost", "pwl_cost"}:
                fl = fl.union(false_elm_links(net, element_type, "element", net[element_type][
                    "et"]))
            elif element_type == "measurement":
                fl = fl.union(false_elm_links(net, element_type, "element", net[element_type][
                    "element_type"]))
            else:
                fl = fl.union(false_elm_links(net, element_type, "bus", "bus"))
            if len(fl):
                false_links[element_type] = fl
    return false_links


def pp_elements(bus=True, bus_elements=True, branch_elements=True, other_elements=True,
                cost_tables=False, res_elements=False):
    """
    Returns a set of pandapower elements.
    """
    pp_elms = set()
    if bus:
        pp_elms |= {"bus"}
        if res_elements:
            pp_elms |= {"res_bus"}
    pp_elms |= set([el[0] for el in element_bus_tuples(
        bus_elements=bus_elements, branch_elements=branch_elements, res_elements=res_elements)])
    if other_elements:
        pp_elms |= {"measurement"}
    if cost_tables:
        pp_elms |= {"poly_cost", "pwl_cost"}
    return pp_elms


def branch_element_bus_dict(include_switch=False, sort=None):
    """
    Returns a dict with keys of branch elements and values of bus column names as list.
    """
    msg = ("The parameter 'sort' is deprecated to function branch_element_bus_dict() with "
           "pp.version >= 2.12. The default was False but the behaviour was changed to True.")
    if sort is not None:
        if Version(__version__) < Version('2.13'):
            warnings.warn(msg, category=DeprecationWarning)
        else:
            raise DeprecationWarning(msg)
    elif Version(__version__) < Version('2.13'):
        logger.debug(msg)

    ebts = element_bus_tuples(bus_elements=False, branch_elements=True, res_elements=False)
    bebd = dict()
    for et, bus in ebts:
        if et in bebd.keys():
            bebd[et].append(bus)
        else:
            bebd[et] = [bus]
    if not include_switch:
        del bebd["switch"]
    return bebd


def element_bus_tuples(bus_elements=True, branch_elements=True, res_elements=False):
    """
    Utility function
    Provides the tuples of elements and corresponding columns for buses they are connected to
    :param bus_elements: whether tuples for bus elements e.g. load, sgen, ... are included
    :param branch_elements: whether branch elements e.g. line, trafo, ... are included
    :param res_elements: whether result table names e.g. res_sgen, res_line, ... are included
    :param return_type: which type the output has
    :return: list of tuples with element names and column names
    """
    if Version(__version__) < Version('2.13'):
        logger.debug("element_bus_tuples() returns a list of tuples instead of a set of tuples "
                     "since pp.version >= 2.12.")
    ebts = list()
    if bus_elements:
        ebts += [("sgen", "bus"), ("load", "bus"), ("ext_grid", "bus"), ("gen", "bus"),
                 ("ward", "bus"), ("xward", "bus"), ("shunt", "bus"),
                 ("storage", "bus"), ("asymmetric_load", "bus"), ("asymmetric_sgen", "bus"),
                 ("motor", "bus")]
    if branch_elements:
        ebts += [("line", "from_bus"), ("line", "to_bus"), ("impedance", "from_bus"),
                 ("impedance", "to_bus"), ("switch", "bus"), ("trafo", "hv_bus"),
                 ("trafo", "lv_bus"), ("trafo3w", "hv_bus"), ("trafo3w", "mv_bus"),
                 ("trafo3w", "lv_bus"), ("dcline", "from_bus"), ("dcline", "to_bus")]
    if res_elements:
        elements_without_res = ["switch", "measurement", "asymmetric_load", "asymmetric_sgen"]
        ebts += [("res_" + ebt[0], ebt[1]) for ebt in ebts if ebt[0] not in elements_without_res]
    return ebts


def count_elements(net, return_empties=False, **kwargs):
    """Counts how much elements of which element type exist in the pandapower net

    Parameters
    ----------
    net : pandapowerNet
        pandapower net
    return_empties : bool, optional
        whether element types should be listed if no element exist, by default False

    Other Parameters
    ----------------
    kwargs : dict[str,bool], optional
        arguments (passed to pp_elements()) to narrow considered element types.
        If nothing is passed, an empty dict is passed to pp_elements(), by default None

    Returns
    -------
    pd.Series
        number of elements per element type existing in the net

    See also
    --------
    count_group_elements

    Examples
    --------
    >>> from pandapower import count_elements
    >>> from pandapower.networks.power_system_test_cases import case9
    >>> count_elements(case9(), bus_elements=False)
    bus     9
    line    9
    dtype: int32
    """
    return pd.Series({et: net[et].shape[0] for et in pp_elements(**kwargs) if return_empties or \
                      bool(net[et].shape[0])}, dtype=np.int64)



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import warnings
from typing_extensions import overload
import pandas as pd

from pandapower.plotting.generic_geodata import create_generic_coordinates
from pandapower.plotting.plotly.traces import create_bus_trace, create_line_trace, \
    create_dcline_trace, create_scale_trace,  create_trafo_trace, draw_traces, \
    _create_node_trace, _create_branch_trace
from pandapower.plotting.plotly.mapbox_plot import *

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging
logger = logging.getLogger(__name__)


def get_hoverinfo(net, element, precision=3, sub_index=None):
    hover_index = net[element].index
    if element == "bus":
        # load_str, sgen_str, vsc_str = [], [], []
        load_str, sgen_str = [], []
        for ln in [net.load.loc[net.load.bus == b, "p_mw"].sum() for b in net.bus.index]:
            load_str.append("Load: {:.3f} MW<br />".format(ln) if ln != 0. else "")
        for s in [net.sgen.loc[net.sgen.bus == b, "p_mw"].sum() for b in net.bus.index]:
            sgen_str.append("Static generation: {:.3f} MW<br />".format(s) if s != 0. else "")
        # we do not really need vsc result for every bus:
        #for vn in [net.res_vsc.loc[net.vsc.bus == b, "p_mw"].fillna(0).sum() for b in net.bus.index]:
        #    vsc_str.append("VSC: {:.3f} MW<br />".format(vn) if vn != 0. else "")
        hoverinfo = (
                "Index: " + net.bus.index.astype(str) + '<br />' +
                "Name: " + net.bus['name'].astype(str) + '<br />' +
                'V_n: ' + net.bus['vn_kv'].round(precision).astype(str) + ' kV' + '<br />' + load_str + sgen_str).tolist()
    elif element == "bus_dc":
        vsc_str = []
        for vn in [net.res_vsc.loc[net.vsc.bus_dc == b, "p_dc_mw"].fillna().sum() for b in net.bus_dc.index]:
            vsc_str.append("VSC: {:.3f} MW<br />".format(vn) if vn != 0. else "")
        hoverinfo = (
                "Index: " + net.bus_dc.index.astype(str) + '<br />' +
                "Name: " + net.bus_dc['name'].astype(str) + '<br />' +
                'V_n: ' + net.bus_dc['vn_kv'].round(precision).astype(str) + ' kV' + '<br />' + vsc_str).tolist()
    elif element == "line":
        hoverinfo = (
                "Index: " + net.line.index.astype(str) + '<br />' +
                "Name: " + net.line['name'].astype(str) + '<br />' +
                'Length: ' + net.line['length_km'].round(precision).astype(str) + ' km' + '<br />' +
                'R: ' + (net.line['length_km'] * net.line['r_ohm_per_km'] / net.line['parallel']).round(precision).astype(str)
                + ' Ohm' + '<br />'
                + 'X: ' + (net.line['length_km'] * net.line['x_ohm_per_km'] / net.line['parallel']).round(precision).astype(str)
                + ' Ohm'
                + net.line['parallel'].apply(lambda x: f'<br />Parallel: {x}' if x > 1 else  '<br />')).tolist()
    elif element == "line_dc":
        hoverinfo = (
                "Index: " + net.line_dc.index.astype(str) + '<br />' +
                "Name: " + net.line_dc['name'].astype(str) + '<br />' +
                'Length: ' + net.line_dc['length_km'].round(precision).astype(str) + ' km' + '<br />' +
                'R: ' + (net.line_dc['length_km'] * net.line['r_ohm_per_km'] / net.line['parallel']).round(precision).astype(str)
                + ' Ohm' + '<br />'
                + net.line_dc['parallel'].apply(lambda x: f'<br />Parallel: {x}' if x > 1 else  '<br />')).tolist()
    elif element == "trafo":
        hoverinfo = (
                "Index: " + net.trafo.index.astype(str) + '<br />' +
                "Name: " + net.trafo['name'].astype(str) + '<br />' +
                'V_n HV: ' + net.trafo['vn_hv_kv'].round(precision).astype(str) + ' kV' + '<br />' +
                'V_n LV: ' + net.trafo['vn_lv_kv'].round(precision).astype(str) + ' kV' + '<br />' +
                'Tap pos.: ' + net.trafo['tap_pos'].astype(str) + '<br />').tolist()
    elif element == "trafo3w":
        hoverinfo = (
                "Index: " + net.trafo3w.index.astype(str) + '<br />' +
                "Name: " + net.trafo3w['name'].astype(str) + '<br />' +
                'V_n HV: ' + net.trafo3w['vn_hv_kv'].round(precision).astype(str) + ' kV' + '<br />' +
                'V_n MV: ' + net.trafo3w['vn_mv_kv'].round(precision).astype(str) + ' kV' + '<br />' +
                'V_n LV: ' + net.trafo3w['vn_lv_kv'].round(precision).astype(str) + ' kV' + '<br />' +
                'Tap pos.: ' + net.trafo3w['tap_pos'].astype(str) + '<br />').tolist()
    elif element == "ext_grid":
        hoverinfo = (
                "Index: " + net.ext_grid.index.astype(str) + '<br />' +
                "Name: " + net.ext_grid['name'].astype(str) + '<br />' +
                'V_m: ' + net.ext_grid['vm_pu'].round(precision).astype(str) + ' p.u.' + '<br />' +
                'V_a: ' + net.ext_grid['va_degree'].round(precision).astype(str) + ' °' + '<br />').tolist()
        hover_index = net.ext_grid.bus.tolist()
    elif element == "vsc":
        hoverinfo = (
                "Index: " + net.vsc.index.astype(str) + '<br />' +
                "Name: " + net.vsc['name'].astype(str) + '<br />' +
                'P: ' + net.res_vsc['p_mw'].fillna(0).round(precision).astype(str) + ' MW' + '<br />' +
                'Q: ' + net.res_vsc['q_mvar'].fillna(0).round(precision).astype(str) + ' MVAr' + '<br />').tolist()
        hover_index = net.vsc.bus.tolist()
    else:
        return None
    hoverinfo = pd.Series(index=hover_index, data=hoverinfo, dtype=object)
    if sub_index is not None:
        hoverinfo = hoverinfo.loc[list(sub_index)]
    return hoverinfo


@overload
def simple_plotly(net, respect_switches=True, use_line_geo=None, on_map=False,
                  *, map_style='basic', figsize=1.0, aspectratio='auto',
                  line_width=1.0, bus_size=10.0, ext_grid_size=20.0,
                  bus_color="blue", line_color='grey', trafo_color='green',
                  trafo3w_color='green', ext_grid_color="yellow",
                  filename='temp-plot.html', auto_open=True, showlegend=True,
                  additional_traces=None, zoomlevel=11, auto_draw_traces=True, hvdc_color='cyan'): ...


@overload
@deprecated("projection is deprecated and will be removed in future versions. geojson should always be WGS84.")
def simple_plotly(net, respect_switches=True, use_line_geo=None, on_map=False,
                  projection='epsg:4326', map_style='basic', figsize=1.0, aspectratio='auto',
                  line_width=1.0, bus_size=10.0, ext_grid_size=20.0,
                  bus_color="blue", line_color='grey', trafo_color='green',
                  trafo3w_color='green', ext_grid_color="yellow",
                  filename='temp-plot.html', auto_open=True, showlegend=True,
                  additional_traces=None, zoomlevel=11, auto_draw_traces=True, hvdc_color='cyan'): ...


def simple_plotly(net, respect_switches=True, use_line_geo=None, on_map=False,
                  projection=None, map_style='basic', figsize=1.0, aspectratio='auto',
                  line_width=1.0, bus_size=10.0, ext_grid_size=20.0,
                  bus_color="blue", line_color='grey', trafo_color='green',
                  trafo3w_color='green', ext_grid_color="yellow",
                  filename='temp-plot.html', auto_open=True, showlegend=True,
                  additional_traces=None, zoomlevel=11, auto_draw_traces=True, hvdc_color='cyan'):
    """
    Plots a pandapower network as simple as possible in plotly.
    If no geodata is available, artificial geodata is generated. For advanced plotting see the tutorial

    INPUT:
        **net** (pandapowerNet) - The pandapower format network.

    OPTIONAL:
        **respect_switches** (bool, True) - Respect switches when artificial geodata is created

        **use_line_geo** (bool, True) - defines if lines patches are based on
        net.line.geo of the lines (True) or on net.bus.geo of the connected buses (False)

        **on_map** (bool, False) - enables using mapbox plot in plotly.
        If provided geodata are not real geo-coordinates in lon/lat form, on_map will be set to False.

        **projection** (String, None) - defines a projection from which network geo-data will be transformed to
        lat-long. For each projection a string can be found at http://spatialreference.org/ref/epsg/

        **map_style** (str, 'basic') - enables using mapbox plot in plotly

            - 'streets'
            - 'bright'
            - 'light'
            - 'dark'
            - 'satellite'

        **figsize** (float, 1) - aspectratio is multiplied by it in order to get final image size

        **aspectratio** (tuple, 'auto') - when 'auto' it preserves original aspect ratio of the network geodata;
        any custom aspectration can be given as a tuple, e.g. (1.2, 1)

        **line_width** (float, 1.0) - width of lines

        **bus_size** (float, 10.0) -  size of buses to plot.

        **ext_grid_size** (float, 20.0) - size of ext_grids to plot.

            See bus sizes for details. Note: ext_grids are plotted as rectangles

        **bus_color** (String, "blue") - Bus Color. Init as first value of color palette.

        **line_color** (String, 'grey') - Line Color. Init is grey

        **trafo_color** (String, 'green') - Trafo Color. Init is green

        **trafo3w_color** (String, 'green') - Trafo 3W Color. Init is blue

        **ext_grid_color** (String, 'yellow') - External Grid Color. Init is yellow

        **auto_open** (bool, True) - automatically open plot in browser

        **showlegend** (bool, True) - If True, a legend will be shown

        **additional_traces** (list, None) - List with additional, user-created traces that will
        be appended to the simple_plotly traces before drawing all traces

        **zoomlevel** (int, 11) - initial mapbox-zoomlevel on a map if `on_map=True`. Small
        values = less zoom / larger area shown

        **auto_draw_traces** (bool, True) - if True, a figure with the drawn traces is returned.
        If False, the traces and a dict with settings is returned

        **hvdc_color** (str, "cyan") - color for HVDC lines

    OUTPUT:
        **figure** (graph_objs._figure.Figure) figure object
    """
    if projection is not None:
        warnings.warn(
            FutureWarning(
                "projection is deprecated and will be removed in future versions. geojson should always be WGS84."
            ),
            stacklevel=2
        )

    settings = dict(
        on_map=on_map,
        map_style=map_style,
        figsize=figsize,
        aspectratio=aspectratio,
        filename=filename,
        auto_open=auto_open,
        showlegend=showlegend,
        zoomlevel=zoomlevel
    )

    traces, _ = _simple_plotly_generic(
        net=net,
        respect_separators=respect_switches,
        use_branch_geodata=use_line_geo,
        branch_width=line_width,
        node_size=bus_size,
        ext_grid_size=ext_grid_size,
        node_color=bus_color,
        branch_color=line_color,
        trafo_color=trafo_color,
        trafo3w_color=trafo3w_color,
        ext_grid_color=ext_grid_color,
        node_element="bus",
        branch_element="line",
        trans_element="trafo",
        trans3w_element="trafo3w",
        branch_trace_func=create_line_trace,
        node_trace_func=create_bus_trace,
        hoverinfo_func=get_hoverinfo,
        hvdc_color=hvdc_color,
        settings=settings
    )
    if additional_traces:
        if isinstance(additional_traces, dict):
            additional_traces = [additional_traces]

        shift = 0
        for weighted_trace in additional_traces:
            # for weighted_marker_traces "meta" should include information for the "scale legend"
            if ("meta" in weighted_trace) and (weighted_trace["meta"]["show_scale_legend"]):
                sc_trace = create_scale_trace(net, weighted_trace, down_shift=shift)
                traces.extend(sc_trace)
                shift += len(weighted_trace["meta"]["scale_marker_size"])

        traces.extend(additional_traces)
    if auto_draw_traces:
        return draw_traces(traces, **settings)
    else:
        return traces, settings


def _simple_plotly_generic(net, respect_separators, use_branch_geodata, branch_width, node_size, ext_grid_size,
                           node_color, branch_color, trafo_color, trafo3w_color, ext_grid_color,
                           node_element, branch_element, trans_element, trans3w_element,
                           branch_trace_func, node_trace_func, hoverinfo_func,
                           hvdc_color="cyan", settings=None, **kwargs):

    if 'projection' in kwargs:
        warnings.warn(
            FutureWarning(
                "projection is deprecated and will not be used. geojson should always be WGS84."
            ),
            stacklevel=2
        )
        kwargs.pop('projection')

    settings_defaults = {  # if no settings are provided, these are used
        'on_map': kwargs.get('on_map', True),
        'map_style': kwargs.get('map_style', 'basic'),
        'figsize': kwargs.get('figsize', 1.),
        'aspectratio': kwargs.get('aspectratio', 'auto'),
        'filename': kwargs.get('filename', 'temp-plot.html'),
        'auto_open': kwargs.get('auto_open', 'auto'),
        'showlegend': kwargs.get('showlegend', True),
        'zoomlevel': kwargs.get('zoomlevel', 11)
    }

    settings = settings_defaults | settings if settings else {}  # add missing settings to settings dict

    if len(net[node_element]["geo"].dropna()) == 0:
        logger.warning(
            "No or insufficient geodata available --> Creating artificial coordinates. This may take some time..."
        )
        create_generic_coordinates(net, respect_switches=respect_separators)
        if settings['on_map']:
            logger.warning(
                "Map plots not available with artificial coordinates and will be disabled!")
            settings['on_map'] = False

    # ----- Nodes (Buses) ------
    # initializing node trace
    hoverinfo = hoverinfo_func(net, element=node_element)
    node_trace = node_trace_func(
        net,
        net[node_element].index,
        size=node_size,
        color=node_color,
        infofunc=hoverinfo
    )
    # ----- branches (Lines) ------
    # if node geodata is available, but no branch geodata
    if use_branch_geodata is None:
        use_branch_geodata = False if len(net[branch_element]["geo"]) == 0 else True
    elif use_branch_geodata and len(net[branch_element]["geo"]) == 0:
        logger.warning(
            "No or insufficient line geodata available --> only bus geodata will be used.")
        use_branch_geodata = False
    hoverinfo = hoverinfo_func(net, element=branch_element)
    branch_traces = branch_trace_func(
        net,
        net[branch_element].index,
        use_branch_geodata,
        respect_separators,
        color=branch_color,
        width=branch_width,
        infofunc=hoverinfo
    )
    trans_trace = []
    trans_trace3w = []
    ext_grid_trace = []
    dc_line_trace = []
    # ----- Trafos ------
    if 'trafo' in net and len(net.trafo):
        hoverinfo = hoverinfo_func(net, element=trans_element)
        trans_trace = create_trafo_trace(
            net,
            color=trafo_color,
            width=branch_width * 5,
            infofunc=hoverinfo,
            use_line_geo=use_branch_geodata
        )
    # ----- 3W Trafos ------
    if 'trafo3w' in net and len(net.trafo3w):
        hoverinfo = hoverinfo_func(net, element=trans3w_element)
        trans_trace3w = create_trafo_trace(
            net, color=trafo3w_color, trafotype='3W',
            width=branch_width * 5,
            trace_name='3W transformers',
            infofunc=hoverinfo,
            use_line_geo=use_branch_geodata
        )
    # ----- Ext grid ------
    # get external grid from _create_node_trace
    if 'ext_grid' in net and len(net.ext_grid):
        marker_type = 'circle' if settings['on_map'] else 'square'  # workaround because doesn't appear on mapbox if square
        hoverinfo = hoverinfo_func(net, element="ext_grid")
        ext_grid_trace = _create_node_trace(
            net,
            nodes=net.ext_grid[node_element],
            size=ext_grid_size,
            patch_type=marker_type,
            color=ext_grid_color,
            infofunc=hoverinfo,
            trace_name='external grid',
            node_element=node_element,
            branch_element=branch_element
        )
    # ----- HVDC lines ------
    if 'dcline' in net and len(net.dcline):
        dc_line_trace = create_dcline_trace(net, color=hvdc_color)

    return branch_traces + trans_trace + trans_trace3w + ext_grid_trace + node_trace + dc_line_trace, settings


if __name__ == '__main__':
    from pandapower.networks import mv_oberrhein
    from pandapower.plotting.plotly.traces import create_weighted_marker_trace

    net = mv_oberrhein()
    net.load.scaling, net.sgen.scaling = 1, 1
    # different markers and sizemodes as examples
    markers_load = create_weighted_marker_trace(net, elm_type="load", color="red",
                                                patch_type="triangle-up", sizemode="area",
                                                marker_scaling=100)
    markers_sgen = create_weighted_marker_trace(net, elm_type="sgen", color="green",
                                                patch_type="circle-open", sizemode="diameter",
                                                marker_scaling=100, scale_marker_size=[0.2, 0.4])

    fig = simple_plotly(
        net,
        bus_size=1,
        aspectratio="original",
        additional_traces=[markers_sgen, markers_load]
    )



# -*- coding: utf-8 -*-

# Copyright 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


"""Defines constants for named column indices to branch matrix.

Some examples of usage, after defining the constants using the line above,
are::

    branch[3, BR_STATUS] = 0              # take branch 4 out of service
    Ploss = branch[:, PF] + branch[:, PT] # compute real power loss vector

The index, name and meaning of each column of the branch matrix is given
below:

columns 0-10 must be included in input matrix (in case file)
    0.  C{F_BUS}       from bus number
    1.  C{T_BUS}       to bus number
    2.  C{BR_R}        resistance (p.u.)
    3.  C{BR_X}        reactance (p.u.)
    4.  C{BR_B}        total line charging susceptance (p.u.)
    5.  C{RATE_A}      MVA rating A (long term rating)
    6.  C{RATE_B}      MVA rating B (short term rating)
    7.  C{RATE_C}      MVA rating C (emergency rating)
    8.  C{TAP}         transformer off nominal turns ratio
    9.  C{SHIFT}       transformer phase shift angle (degrees)
    10. C{BR_STATUS}   initial branch status, 1 - in service, 0 - out of service
    11. C{ANGMIN}      minimum angle difference, angle(Vf) - angle(Vt) (degrees)
    12. C{ANGMAX}      maximum angle difference, angle(Vf) - angle(Vt) (degrees)

columns 13-16 are added to matrix after power flow or OPF solution
they are typically not present in the input matrix
    13. C{PF}          real power injected at "from" bus end (MW)
    14. C{QF}          reactive power injected at "from" bus end (MVAr)
    15. C{PT}          real power injected at "to" bus end (MW)
    16. C{QT}          reactive power injected at "to" bus end (MVAr)

columns 17-18 are added to matrix after OPF solution
they are typically not present in the input matrix

(assume OPF objective function has units, C{u})
    17. C{MU_SF}       Kuhn-Tucker multiplier on MVA limit at "from" bus (u/MVA)
    18. C{MU_ST}       Kuhn-Tucker multiplier on MVA limit at "to" bus (u/MVA)

columns 19-20 are added to matrix after SCOPF solution
they are typically not present in the input matrix

(assume OPF objective function has units, C{u})
    19. C{MU_ANGMIN}   Kuhn-Tucker multiplier lower angle difference limit
    20. C{MU_ANGMAX}   Kuhn-Tucker multiplier upper angle difference limit

@author: Ray Zimmerman (PSERC Cornell)
@author: Richard Lincoln
"""

# define the indices
DC_F_BUS       = 0    # f, from bus number
DC_T_BUS       = 1    # t, to bus number
DC_BR_R        = 2    # r, resistance (p.u.)
DC_BR_G        = 3    # b, total line charging susceptance (p.u.)  # todo Roman check if necessary
DC_RATE_A      = 4    # rateA, MVA rating A (long term rating)
DC_RATE_B      = 5    # rateB, MVA rating B (short term rating)
DC_RATE_C      = 6    # rateC, MVA rating C (emergency rating)
DC_BR_STATUS   = 7   # initial branch status, 1 - in service, 0 - out of service

# included in power flow solution, not necessarily in input
DC_PF          = 8   # real power injected at "from" bus end (MW)
DC_IF          = 9   # current injected at "from" bus end (p.u.)
DC_PT          = 10   # real power injected at "to" bus end (MW)
DC_IT          = 11   # current injected at "to" bus end (p.u.)

DC_BR_R_ASYM = 12   # todo Roman check if necessary
DC_BR_X_ASYM = 13   # todo Roman check if necessary

DC_TDPF = 14  ### TDPF not implemented for DC lines

branch_dc_cols = 15
